""" Main namespace commands. """

import click

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import output_generic, output_service
from gandi.cli.core.params import pass_gandi


@cli.command()
@pass_gandi
def setup(gandi):
    """ Initialize Gandi CLI configuration.

    Create global configuration directory with API credentials
    """
    intro = """Welcome to GandiCLI, let's configure a few things before we \
start.
"""

    outro = """
Setup completed. You can now:
* use 'gandi' to see all command.
* use 'gandi vm create' to create and access a Virtual Machine.
* use 'gandi paas create' to create and access a SimpleHosting instance.
"""
    gandi.echo(intro)
    gandi.init_config()
    gandi.echo(outro)


@cli.command()
@pass_gandi
def api(gandi):
    """Display information about API used."""
    key_name = 'API version'

    result = gandi.api.info()
    result[key_name] = result.pop('api_version')
    output_generic(gandi, result, [key_name])

    return result


@cli.command()
@click.argument('command', required=False, nargs=-1)
@click.pass_context
def help(ctx, command):
    """Display help for a command."""
    command = ' '.join(command)
    if not command:
        click.echo(cli.get_help(ctx))
        return

    cmd = cli.get_command(ctx, command)
    if cmd:
        click.echo(cmd.get_help(ctx))
    else:
        click.echo(cli.get_help(ctx))


@cli.command()
@click.argument('service', required=False)
@pass_gandi
def status(gandi, service):
    """Display current status from status.gandi.net."""

    if not service:
        global_status = gandi.status.status()
        if global_status['status'] == 'FOGGY':
            # something is going on but not affecting services
            filters = {
                'category': 'Incident',
                'current': True,
            }
            events = gandi.status.events(filters)
            for event in events:
                if event['services']:
                    # do not process services
                    continue
                event_url = gandi.status.event_timeline(event)
                service_detail = '%s - %s' % (event['title'], event_url)
                gandi.echo(service_detail)

    # then check other services
    descs = gandi.status.descriptions()
    needed = services = gandi.status.services()
    if service:
        needed = [serv for serv in services
                  if serv['name'].lower() == service.lower()]

    for serv in needed:
        if serv['status'] != 'STORMY':
            output_service(gandi, serv['name'], descs[serv['status']])
            continue

        filters = {
            'category': 'Incident',
            'services': serv['name'],
            'current': True,
        }
        events = gandi.status.events(filters)
        for event in events:
            event_url = gandi.status.event_timeline(event)
            service_detail = '%s - %s' % (event['title'], event_url)
            output_service(gandi, serv['name'], service_detail)

    return services



# -*- coding: utf-8 -*-

from .base import CommandTestCase
from gandi.cli.commands import oper


class OperTestCase(CommandTestCase):

    def test_list(self):

        result = self.invoke_with_exceptions(oper.list, [])
        self.assertEqual(result.output, """\
id        : 99002
type      : hosting_migration_vm
step      : RUN
----------
id        : 99001
type      : hosting_migration_vm
step      : RUN
----------
id        : 100303
type      : certificate_update
step      : WAIT
----------
id        : 100302
type      : certificate_update
step      : RUN
----------
id        : 100300
type      : certificate_update
step      : RUN
----------
id        : 100200
type      : billing_prepaid_add_money
step      : BILL
----------
id        : 100100
type      : domain_renew
step      : BILL
""")
        self.assertEqual(result.exit_code, 0)

    def test_info(self):

        result = self.invoke_with_exceptions(oper.info, ['100100'])

        self.assertEqual(result.output, """\
id        : 100100
type      : domain_renew
step      : BILL
last_error:
""")
        self.assertEqual(result.exit_code, 0)




try:
    from unittest import mock
except ImportError:
    import mock


try:
    import unittest2 as unittest
except ImportError:
    import unittest


try:
    import configparser as ConfigParser
except ImportError:
    import ConfigParser


try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


try:
    from cStringIO import StringIO as ReasonableBytesIO
except ImportError:
    from io import BytesIO as ReasonableBytesIO



""" Custom command line validation parameters. """

import re
import sys

import click
from click.decorators import _param_memo

from gandi.cli.core.base import GandiContextHelper


class GandiChoice(click.Choice):

    """ Base class for custom Choice parameters. """

    gandi = None

    def __init__(self):
        """ Initialize choices list. """
        self._choices = []
        self.case_sensitive = None

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        raise NotImplementedError

    @property
    def choices(self):
        """ Retrieve choices from API if possible"""
        if not self._choices:
            gandi = self.gandi or GandiContextHelper()
            self._choices = self._get_choices(gandi)
            if not self._choices:
                api = gandi.get_api_connector()
                gandi.echo('Please check that you are connecting to the good '
                           "api '%s' and that it's running." % (api.host))
                sys.exit(1)

        return self._choices

    def convert(self, value, param, ctx):
        """ Internal method to use correct context. """
        self.gandi = ctx.obj
        return click.Choice.convert(self, value, param, ctx)

    def convert_deprecated_value(self, value):
        """ To override when needed """
        return value


class DatacenterParamType(GandiChoice):

    """ Choice parameter to select an available datacenter. """

    name = 'datacenter'

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        iso_codes = []
        dc_codes = []
        for item in gandi.datacenter.list():
            iso_codes.append(item['iso'])
            if item.get('dc_code'):
                dc_codes.append(item['dc_code'])
        return dc_codes + iso_codes

    def convert(self, value, param, ctx):
        """ Convert value to uppercase. """
        self.gandi = ctx.obj
        value = value.upper()
        return click.Choice.convert(self, value, param, ctx)

    def convert_deprecated_value(self, value):
        """ To update the configuration with the new datacenter naming """
        convert = {
            'FR': 'FR-SD2',
            'LU': 'LU-BI1',
            'US': 'US-BA1',
        }
        return convert.get(value, value)


class PaasTypeParamType(GandiChoice):

    """ Choice parameter to select an available PaaS instance type. """

    name = 'paas type'

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        return [item['name'] for item in gandi.paas.type_list()]


class IntChoice(click.Choice):

    """ Choice parameter to select an integer value in a set of int values."""

    name = 'integer choice'

    def convert(self, value, param, ctx):
        """ Convert value to int. """
        self.gandi = ctx.obj
        try:
            value = str(value)
        except Exception:
            pass
        value = click.Choice.convert(self, value, param, ctx)
        return int(value)


class DiskImageParamType(GandiChoice):

    """ Choice parameter to select an available disk image. """

    name = 'images'

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        image_list = []
        for item in gandi.image.list():
            label = item['label']
            if item['visibility'] == 'deprecated':
                label = '*%s' % label
            image_list.append(label)
        disk_list = [item['name'] for item in gandi.disk.list_create()]
        return sorted(tuple(set(image_list))) + disk_list

    def convert(self, value, param, ctx):
        """ Try to find correct disk image regarding version. """
        self.gandi = ctx.obj
        # remove deprecated * prefix
        choices = [choice.replace('*', '') for choice in self.choices]
        value = value.replace('*', '')
        # Exact match
        if value in choices:
            return value

        # Try to find 64 bits version
        new_value = '%s 64 bits' % value
        if new_value in choices:
            return new_value

        # Try to find without specific bits version
        p = re.compile(' (64|32) bits')
        new_value = p.sub('', value)
        if new_value in choices:
            return new_value

        self.fail('invalid choice: %s. (choose from %s)' %
                  (value, ', '.join(self.choices)), param, ctx)


class KernelParamType(GandiChoice):

    """ Choice parameter to select an available kernel. """

    name = 'kernels'

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        kernel_families = list(gandi.kernel.list().values())
        return [kernel for klist in kernel_families for kernel in klist]

    def convert(self, value, param, ctx):
        """ Try to find correct kernel regarding version. """
        self.gandi = ctx.obj
        # Exact match first
        if value in self.choices:
            return value

        # Also try with x86-64 suffix
        new_value = '%s-x86_64' % value
        if new_value in self.choices:
            return new_value

        self.fail('invalid choice: %s. (choose from %s)' %
                  (value, ', '.join(self.choices)), param, ctx)


class SnapshotParamType(GandiChoice):

    """ Choice parameter to select an available snapshot profile. """

    name = 'snapshot profile'
    target = None

    def __init__(self, target=None):
        """ Initialize choices list. """
        self._choices = []
        self.target = target
        self.case_sensitive = None

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        return [str(item['id'])
                for item in gandi.snapshotprofile.list(target=self.target)]

    def convert(self, value, param, ctx):
        """ Convert value to int. """
        self.gandi = ctx.obj
        value = click.Choice.convert(self, value, param, ctx)
        return int(value)


class CertificatePackage(GandiChoice):

    """ Choice parameter to select an available certificate package. """

    name = 'certificate package'

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        return [item['name'] for item in gandi.certificate.package_list()]


class CertificatePackageType(CertificatePackage):

    """ Choice parameter to select an available certificate package type. """

    name = 'certificate package type'

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        packages = super(CertificatePackageType, self)._get_choices(gandi)
        return list(set([pack.split('_')[1] for pack in packages]))


class CertificatePackageMax(CertificatePackage):

    """
    Choice parameter to select an available certificate package max altname.
    """

    name = 'certificate package max'

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        packages = super(CertificatePackageMax, self)._get_choices(gandi)
        ret = list(set([pack.split('_')[2] for pack in packages]))
        if 'w' in ret:
            ret.remove('w')
        return ret

    def convert(self, value, param, ctx):
        """ Convert value to int. """
        self.gandi = ctx.obj
        value = click.Choice.convert(self, value, param, ctx)
        return int(value)


class CertificatePackageWarranty(CertificatePackage):

    """ Choice parameter to select an available certificate warranty. """

    name = 'certificate package warranty'

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        packages = super(CertificatePackageWarranty, self)._get_choices(gandi)
        return list(set([pack.split('_')[3] for pack in packages]))

    def convert(self, value, param, ctx):
        """ Convert value to int. """
        self.gandi = ctx.obj
        value = click.Choice.convert(self, value, param, ctx)
        return int(value)


class CertificateDcvMethod(click.Choice):

    """ Choice parameter to select a certificate dcv method.

    * 'email' will send you an email to check domain ownership
    * 'dns' will require you to add a TXT record in your domain zone
    * 'file' will require you to add a file on you server
    * 'auto' can only be used when your domain and its zone are on the
       same gandi account you are currently using (gandi will add the TXT
       dns record).
    """

    name = 'certificate dcv method'
    choices = ['email', 'dns', 'file', 'auto']

    def __init__(self):
        """ Initialize choices list. """
        pass


class IpType(click.Choice):
    """ Choice parameter to filter on ip types.

    * 'private' will only retrieve private ips
    * 'public' will only retrieve public ips
    """

    name = 'ip type'
    choices = ['private', 'public']

    def __init__(self):
        """ Initialize choices list. """
        pass


class StringConstraint(click.types.StringParamType):

    """ Check that provided string matches constraints."""

    name = 'string constraints'

    def __init__(self, minlen=None, maxlen=None):
        self.min = minlen
        self.max = maxlen

    def convert(self, value, param, ctx):
        value = click.types.StringParamType.convert(self, value, param, ctx)
        rv = len(value)
        if self.min is not None and rv < self.min or \
           self.max is not None and rv > self.max:
            if self.min is None:
                self.fail('%s is longer than the maximum valid length '
                          '%s.' % (rv, self.max), param, ctx)
            elif self.max is None:
                self.fail('%s is shorter than the minimum valid length '
                          '%s.' % (rv, self.min), param, ctx)
            else:
                self.fail('%s is not in the valid length range of %s to %s.'
                          % (rv, self.min, self.max), param, ctx)
        return value

    def __repr__(self):
        return 'StringConstraint(%r, %r)' % (self.min, self.max)


class EmailParamType(click.ParamType):

    """Check the email value and return a list ['login', 'domain']. """

    name = 'email'

    def convert(self, value, param, ctx):
        """ Validate value using regexp. """
        rxp = '^[^@]+?@[-.a-z0-9]+$'
        regex = re.compile(rxp, re.I)
        try:
            if regex.match(value):
                value = value.split("@")
                return value
            else:
                self.fail('%s is not a valid email address' % value, param,
                          ctx)
        except ValueError:
            self.fail('%s is not a valid email address' % value, param, ctx)


class SizeParamType(click.ParamType):
    name = 'size'
    suffixes = {'M': 0,
                'G': 1,
                'T': 2}
    prefixes = ['+']

    def convert(self, value, param, ctx):
        prefix = ''
        suffix = ''
        for i, c in enumerate(value):
            if not c.isdigit():
                if c in self.prefixes:
                    prefix = c
                    continue
                suffix = value[i:]
                value = value[:i]
                break
        try:
            mul = self.suffixes[suffix] if suffix else 0
            return prefix, int(value) * (1 << (mul * 10))
        except ValueError:
            self.fail("%r is not an integer" % (value))
        except KeyError:
            self.fail("%r is not a supported suffix" % (suffix))


class BackendParamType(click.ParamType):

    """

    Check the validity of the server ip and port and return a dict
    ['ip', 'port'].

    """

    name = 'backend'

    def convert(self, value, param, ctx):
        """ Validate value using regexp. """
        rxp = "^(((([1]?\d)?\d|2[0-4]\d|25[0-5])\.){3}(([1]?\d)?\d|2[0-4]\d|"\
              "25[0-5]))|([\da-fA-F]{1,4}(\:[\da-fA-F]{1,4}){7})|(([\da-fA-F]"\
              "{1,4}:){0,5}::([\da-fA-F]{1,4}:)"\
              "{0,5}[\da-fA-F]{1,4})$"
        regex = re.compile(rxp, re.I)
        backend = {}
        if value.count(':') == 0:
            # port is not set
            backend['ip'] = value
        elif value.count(':') == 7:
            # it's an ipv6 without port
            backend['ip'] = value
        elif value.count(':') == 8:
            # it's an ipv6 with port
            backend['ip'] = value.rsplit(':', 1)[0]
            backend['port'] = int(value.rsplit(':', 1)[1])
        else:
            backend['ip'] = value.split(':')[0]
            backend['port'] = int(value.split(':')[1])
        try:
            if regex.match(backend['ip']):
                return backend
            else:
                self.fail('%s is not a valid ip address' %
                          backend['ip'], param, ctx)
        except ValueError:
            self.fail('%s is not a valid ip address' % backend['ip'], param,
                      ctx)


class WebAccNameParamType(GandiChoice):
    """ Choice a webaccelerator """

    name = 'webacc list'

    def _get_choices(self, gandi):
        """ Internal method to get choice list """
        return [str(item['name']) for item in gandi.webacc.list()]


class WebAccVhostParamType(GandiChoice):
    """ Retrieve vhost on a webaccelerator """
    name = 'webacc vhost list'

    def _get_choices(self, gandi):
        """ Internal method to get choice list """
        return [str(item['name']) for item in gandi.webacc.vhost_list()]


class OperStepParamType(click.Choice):
    """ Choice parameter to filter on operation step """

    name = 'oper step'
    choices = ['BILL', 'WAIT', 'RUN', 'ERROR']

    def __init__(self):
        """ Initialize choices list. """
        pass


class DNSRecordsParamType(GandiChoice):

    """ Choice parameter to select a DNS record type. """

    name = 'dns record'

    def _get_choices(self, gandi):
        """ Internal method to get choices list """
        return [item for item in gandi.dns.type_list()]

    def convert(self, value, param, ctx):
        """ Convert value to uppercase. """
        self.gandi = ctx.obj
        value = value.upper()
        return click.Choice.convert(self, value, param, ctx)


class AlgorithmType(click.Choice):
    """ Choice parameter to create DNSSEC key """

    name = 'algorithm'
    choices = ['1', '2', '3', '5', '6', '7', '8', '10', '12', '13', '14', '15',
               '16', '253', '254']

    def __init__(self):
        """ Initialize choices list. """
        pass


class FlagsType(click.Choice):
    """ Choice parameter to create DNSSEC key """

    name = 'flags'
    choices = ['256', '257']

    def __init__(self):
        """ Initialize choices list. """
        pass


DATACENTER = DatacenterParamType()
PAAS_TYPE = PaasTypeParamType()
DISK_IMAGE = DiskImageParamType()
DISK_MAXLIST = 500
KERNEL = KernelParamType()
SNAPSHOTPROFILE_PAAS = SnapshotParamType('paas')
SNAPSHOTPROFILE_VM = SnapshotParamType('vm')
CERTIFICATE_PACKAGE = CertificatePackage()
CERTIFICATE_PACKAGE_TYPE = CertificatePackageType()
CERTIFICATE_PACKAGE_MAX = CertificatePackageMax()
CERTIFICATE_PACKAGE_WARRANTY = CertificatePackageWarranty()
CERTIFICATE_DCV_METHOD = CertificateDcvMethod()
EMAIL_TYPE = EmailParamType()
IP_TYPE = IpType()
SIZE = SizeParamType()
BACKEND = BackendParamType()
WEBACC_NAME = WebAccNameParamType()
WEBACC_VHOST_NAME = WebAccVhostParamType()
OPER_STEP = OperStepParamType()
DNS_RECORDS = DNSRecordsParamType()
DNSSEC_ALGORITHM = AlgorithmType()
DNSSEC_FLAGS = FlagsType()


class GandiOption(click.Option):

    """ Custom command option class for handling configuration files.

    When no value was found on command line, try to pull it from configuration
    Display default or configuration value when needed
    """

    def display_value(self, ctx, value):
        """ Display value to be used for this parameter. """
        gandi = ctx.obj
        gandi.log('%s: %s' % (self.name, (value if value is not None
                                          else 'Not found')))

    def get_default(self, ctx):
        """ Retrieve default value and display it when prompt disabled. """
        value = click.Option.get_default(self, ctx)
        if not self.prompt:
            # value found in default display it
            self.display_value(ctx, value)
        return value

    def consume_value(self, ctx, opts):
        """ Retrieve default value and display it when prompt is disabled. """
        value = click.Option.consume_value(self, ctx, opts)
        if not value:
            # value not found by click on command line
            # now check using our context helper in order into
            # local configuration
            # global configuration
            gandi = ctx.obj
            value = gandi.get(self.name)
            if value is not None:
                # value found in configuration display it
                self.display_value(ctx, value)
            else:
                if self.default is None and self.required:
                    metavar = ''
                    if self.type.name not in ['integer', 'text']:
                        metavar = self.make_metavar()
                    prompt = '%s %s' % (self.help, metavar)
                    gandi.echo(prompt)
        return value

    def handle_parse_result(self, ctx, opts, args):
        """ Save value for this option in configuration
        if key/value pair doesn't already exist.
        Or old value in config was deprecated
        it needs to be updated to the new value format
        but the value keeps the same "meaning"
        """
        gandi = ctx.obj
        needs_update = False
        value, args = click.Option.handle_parse_result(self, ctx, opts, args)

        if value is not None:
            previous_value = gandi.get(global_=True, key=self.name)
            if isinstance(self.type, GandiChoice):
                if value == previous_value:
                    needs_update = True
                value = self.type.convert_deprecated_value(value)
            if not previous_value or needs_update:
                gandi.configure(global_=True, key=self.name, val=value)
        opts[self.name] = value
        value, args = click.Option.handle_parse_result(self, ctx, opts, args)
        return value, args


def option(*param_decls, **attrs):
    """Attach an option to the command.

    All positional arguments are passed as parameter declarations
    to :class:`Option`, all keyword arguments are forwarded unchanged.
    This is equivalent to creating an :class:`Option` instance manually and
    attaching it to the :attr:`Command.params` list.
    """
    def decorator(f):
        _param_memo(f, GandiOption(param_decls, **attrs))
        return f
    return decorator


# create a decorator to pass the Gandi object as context to click calls
pass_gandi = click.make_pass_decorator(GandiContextHelper, ensure=True)



import os
import subprocess

from gandi.cli.core.base import GandiModule
from gandi.cli.modules.iaas import Iaas
from gandi.cli.core.utils import unixpipe


class Docker(GandiModule):
    """
    Module to handle docker vms.

    $ gandi docker create
    $ gandi docker help
    $ gandi docker ps
    $ gandi docker <docker_cmd>

    Note that you can use a per-project docker vm by using
    a local directory gandi configuration using:

    $ gandi config set dockervm foobar

    Or override the current global vm using:

    $ gandi docker --vm bar ps
    """

    @classmethod
    def handle(cls, vm, args):
        """
        Setup forwarding connection to given VM and pipe docker cmds over SSH.
        """
        docker = Iaas.info(vm)
        if not docker:
            raise Exception('docker vm %s not found' % vm)

        if docker['state'] != 'running':
            Iaas.start(vm)

        # XXX
        remote_addr = docker['ifaces'][0]['ips'][0]['ip']

        port = unixpipe.setup(remote_addr, 'root', '/var/run/docker.sock')

        os.environ['DOCKER_HOST'] = 'tcp://localhost:%d' % port
        cls.echo('using DOCKER_HOST=%s' % os.environ['DOCKER_HOST'])

        subprocess.call(['docker'] + list(args))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .base import CommandTestCase
from gandi.cli.commands import forward


class ForwardTestCase(CommandTestCase):

    def test_list(self):

        result = self.invoke_with_exceptions(forward.list, ['iheartcli.com'])

        self.assertEqual(result.output, """\
admin         : admin@cli.sexy
admin         : grumpy@cat.lol
contact       : contact@cli.sexy
""")

        self.assertEqual(result.exit_code, 0)

    def test_create(self):
        args = ['backup@iheartcli.com', '--destination', 'backup@cat.lol']
        result = self.invoke_with_exceptions(forward.create, args)

        self.assertEqual(result.output, """\
Creating mail forward backup@iheartcli.com
""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['domain.forward.create'][0][2]
        self.assertEqual(params['destinations'], ['backup@cat.lol'])

    def test_delete_force(self):
        args = ['admin@iheartcli.com', '--force']
        result = self.invoke_with_exceptions(forward.delete, args)

        self.assertEqual(result.output, '')

        self.assertEqual(result.exit_code, 0)

    def test_delete(self):
        args = ['admin@iheartcli.com']
        result = self.invoke_with_exceptions(forward.delete, args, input='y\n')

        self.assertEqual(result.output, """\
Are you sure to delete the domain mail forward admin@iheartcli.com ? [y/N]: y
""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_force_refused(self):
        args = ['admin@iheartcli.com']
        result = self.invoke_with_exceptions(forward.delete, args, input='\n')

        self.assertEqual(result.output, """\
Are you sure to delete the domain mail forward admin@iheartcli.com ? [y/N]: \
\n""")

        self.assertEqual(result.exit_code, 0)

    def test_update(self):
        args = ['admin@iheartcli.com',
                '--dest-add', 'doge@iheartcli.com',
                '--dest-del', 'grumpy@cat.lol']

        result = self.invoke_with_exceptions(forward.update, args)

        self.assertEqual(result.output, """\
Updating mail forward admin@iheartcli.com
""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['domain.forward.update'][0][2]
        self.assertEqual(params['destinations'], ['admin@cli.sexy',
                                                  'doge@iheartcli.com'])

    def test_update_no_param(self):
        args = ['admin@iheartcli.com']

        result = self.invoke_with_exceptions(forward.update, args)

        self.assertEqual(result.output, """\
Nothing to update: you must provide destinations to update, use \
--dest-add/-a or -dest-del/-d parameters.
""")

        self.assertEqual(result.exit_code, 0)



import re

from .base import CommandTestCase
from ..fixtures.mocks import MockObject
from gandi.cli.core.base import GandiModule
from gandi.cli.commands import contact


class ContactTestCase(CommandTestCase):

    mocks = [('gandi.cli.commands.contact.webbrowser.open',
              MockObject.blank_func)]

    def test_create_dry_run_ok(self):

        args = []
        inputs = ('0\nPeter\nParker\npeter.parker@spiderman.org\n'
                  'Central Park\n2600\nNew York\nUSA\n555-123-456\n'
                  'plokiploki\nplokiploki\n+011.555123456\napikey0001\n')
        result = self.invoke_with_exceptions(contact.create, args,
                                             input=inputs)
        self.assertEqual(re.sub(r'\[\d+\]', '[1234567890123456]',
                                result.output.strip()), """\
Choose your contact type
0- individual
1- company
2- association
3- public body
4- reseller
: 0
What is your first name: Peter
What is your last name: Parker
What is your email address: peter.parker@spiderman.org
What is your street address: Central Park
What is your zipcode: 2600
Which city: New York
Which country: USA
What is your telephone number: 555-123-456
Please enter your password [1234567890123456]: \
\nRepeat for confirmation: \
\nphone: string '555-123-456' does not match '^\\+\\d{1,3}\\.\\d+$'
What is your telephone number: +011.555123456
Please activate you public api access from gandi website, and get the apikey.
Your handle is PP0000-GANDI, and the password is the one you defined.
What is your production apikey: apikey0001
You already have an apikey defined, if you want to use the newly created \
contact, use the env var : \
\nexport API_KEY=apikey0001""")
        self.assertEqual(result.exit_code, 0)

    def test_create_dry_run_unknown_ok(self):

        args = []
        inputs = ('0\nPeter\nParker\ngreen.goblin@spiderman.org\n'
                  'Central Park\n2600\nNew York\nUSA\n555-123-456\n'
                  'plokiploki\nplokiploki\n+011.555123456\napikey0001\n')
        result = self.invoke_with_exceptions(contact.create, args,
                                             input=inputs)
        self.assertEqual(re.sub(r'\[\d+\]', '[1234567890123456]',
                                result.output.strip()), """\
Choose your contact type
0- individual
1- company
2- association
3- public body
4- reseller
: 0
What is your first name: Peter
What is your last name: Parker
What is your email address: green.goblin@spiderman.org
What is your street address: Central Park
What is your zipcode: 2600
Which city: New York
Which country: USA
What is your telephone number: 555-123-456
Please enter your password [1234567890123456]: \
\nRepeat for confirmation: \
\nphone: string '555-123-456' does not match '^\\+\\d{1,3}\\.\\d+$'
What is your telephone number: +011.555123456
planet: Pluto not in list Sun, Mercury, Venus, Earth, Mars, Jupiter, Saturn, \
Uranus, Neptune""")
        self.assertEqual(result.exit_code, 0)

    def test_create_ok(self):

        args = []
        inputs = ('0\nPeter\nParker\npeter.parker@spiderman.org\n'
                  'Central Park\n2600\nNew York\nUSA\n+011.555123456\n'
                  'plokiploki\nplokiploki\napikey0001\n')
        result = self.invoke_with_exceptions(contact.create, args,
                                             input=inputs)
        self.assertEqual(re.sub(r'\[\d+\]', '[1234567890123456]',
                                result.output.strip()), """\
Choose your contact type
0- individual
1- company
2- association
3- public body
4- reseller
: 0
What is your first name: Peter
What is your last name: Parker
What is your email address: peter.parker@spiderman.org
What is your street address: Central Park
What is your zipcode: 2600
Which city: New York
Which country: USA
What is your telephone number: +011.555123456
Please enter your password [1234567890123456]: \
\nRepeat for confirmation: \
\nPlease activate you public api access from gandi website, and get the apikey.
Your handle is PP0000-GANDI, and the password is the one you defined.
What is your production apikey: apikey0001
You already have an apikey defined, if you want to use the newly created \
contact, use the env var : \
\nexport API_KEY=apikey0001""")
        self.assertEqual(result.exit_code, 0)

    def test_create_apikey_ok(self):

        args = []
        inputs = ('0\nPeter\nParker\npeter.parker@spiderman.org\n'
                  'Central Park\n2600\nNew York\nUSA\n+011.555123456\n'
                  'plokiploki\nplokiploki\napikey0002\n')

        GandiModule._conffiles = {'global': {}}

        result = self.invoke_with_exceptions(contact.create, args,
                                             input=inputs)
        self.assertEqual(re.sub(r'\[\d+\]', '[1234567890123456]',
                                result.output.strip()), """\
Choose your contact type
0- individual
1- company
2- association
3- public body
4- reseller
: 0
What is your first name: Peter
What is your last name: Parker
What is your email address: peter.parker@spiderman.org
What is your street address: Central Park
What is your zipcode: 2600
Which city: New York
Which country: USA
What is your telephone number: +011.555123456
Please enter your password [1234567890123456]: \
\nRepeat for confirmation: \
\nPlease activate you public api access from gandi website, and get the apikey.
Your handle is PP0000-GANDI, and the password is the one you defined.
What is your production apikey: apikey0002
Will save your apikey into the config file.""")
        self.assertEqual(result.exit_code, 0)

        self.assertTrue('api' in GandiModule._conffiles['global'])

        api_key = GandiModule._conffiles['global']['api'].get('key')
        self.assertEqual(api_key, 'apikey0002')



import re

from click.exceptions import ClickException

from .base import CommandTestCase
from gandi.cli.commands import disk
from gandi.cli.core.utils.size import disk_check_size
from gandi.cli.core.base import GandiContextHelper


class DiskTestCase(CommandTestCase):

    def test_list(self):

        result = self.invoke_with_exceptions(disk.list, [])

        self.assertEqual(result.output, """name      : sys_1426759833
state     : created
size      : 3072
----------
name      : sys_server01
state     : created
size      : 3072
----------
name      : data
state     : created
size      : 3072
----------
name      : snaptest
state     : created
size      : 3072
----------
name      : newdisk
state     : created
size      : 3072
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_vm(self):

        result = self.invoke_with_exceptions(disk.list, ['--vm'])

        self.assertEqual(result.output, """name      : sys_1426759833
state     : created
size      : 3072
vm        : vm1426759833
----------
name      : sys_server01
state     : created
size      : 3072
vm        : server01
----------
name      : data
state     : created
size      : 3072
vm        : server01
----------
name      : snaptest
state     : created
size      : 3072
----------
name      : newdisk
state     : created
size      : 3072
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_id(self):

        result = self.invoke_with_exceptions(disk.list, ['--id'])

        self.assertEqual(result.output, """name      : sys_1426759833
state     : created
size      : 3072
id        : 4969232
----------
name      : sys_server01
state     : created
size      : 3072
id        : 4969249
----------
name      : data
state     : created
size      : 3072
id        : 4970079
----------
name      : snaptest
state     : created
size      : 3072
id        : 663497
----------
name      : newdisk
state     : created
size      : 3072
id        : 4969233
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_type(self):

        result = self.invoke_with_exceptions(disk.list, ['--type'])

        self.assertEqual(result.output, """name      : sys_1426759833
state     : created
size      : 3072
type      : data
----------
name      : sys_server01
state     : created
size      : 3072
type      : data
----------
name      : data
state     : created
size      : 3072
type      : data
----------
name      : snaptest
state     : created
size      : 3072
type      : snapshot
----------
name      : newdisk
state     : created
size      : 3072
type      : data
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_only_data(self):

        result = self.invoke_with_exceptions(disk.list, ['--only-data'])

        self.assertEqual(result.output, """name      : sys_1426759833
state     : created
size      : 3072
----------
name      : sys_server01
state     : created
size      : 3072
----------
name      : data
state     : created
size      : 3072
----------
name      : newdisk
state     : created
size      : 3072
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_only_snapshot(self):

        result = self.invoke_with_exceptions(disk.list, ['--only-snapshot'])

        self.assertEqual(result.output, """name      : snaptest
state     : created
size      : 3072
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_snapshotprofile(self):

        result = self.invoke_with_exceptions(disk.list, ['--snapshotprofile'])

        self.assertEqual(result.output, """name      : sys_1426759833
state     : created
size      : 3072
----------
name      : sys_server01
state     : created
size      : 3072
----------
name      : data
state     : created
size      : 3072
profile   : minimal
----------
name      : snaptest
state     : created
size      : 3072
----------
name      : newdisk
state     : created
size      : 3072
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_filter_datacenter(self):
        args = ['--datacenter', 'LU-BI1']
        result = self.invoke_with_exceptions(disk.list, args)

        self.assertEqual(result.output, """\
name      : sys_1426759833
state     : created
size      : 3072
----------
name      : newdisk
state     : created
size      : 3072
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_attached(self):

        result = self.invoke_with_exceptions(disk.list, ['--attached'])

        self.assertEqual(result.output, """\
name      : sys_1426759833
state     : created
size      : 3072
----------
name      : sys_server01
state     : created
size      : 3072
----------
name      : data
state     : created
size      : 3072
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_detached(self):

        result = self.invoke_with_exceptions(disk.list, ['--detached'])

        self.assertEqual(result.output, """\
name      : snaptest
state     : created
size      : 3072
----------
name      : newdisk
state     : created
size      : 3072
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_attached_detached_ko(self):
        args = ['--detached', '--attached']
        result = self.invoke_with_exceptions(disk.list, args)

        self.assertEqual(result.output, """\
Usage: list [OPTIONS]

Error: You cannot use both --attached and --detached.
""")

        self.assertEqual(result.exit_code, 2)

    def test_info(self):
        result = self.invoke_with_exceptions(disk.info, ['sys_server01'])

        self.assertEqual(result.output, """name      : sys_server01
state     : created
size      : 3072
type      : data
id        : 4969249
kernel    : 3.12-x86_64 (hvm)
cmdline   : root=/dev/sda ro nosep console=ttyS0
datacenter: FR-SD2
vm        : server01
""")
        self.assertEqual(result.exit_code, 0)

    def test_info_multiple(self):
        args = ['sys_server01', 'data']
        result = self.invoke_with_exceptions(disk.info, args)

        self.assertEqual(result.output, """name      : data
state     : created
size      : 3072
type      : data
id        : 4970079
datacenter: FR-SD2
vm        : server01
----------
name      : sys_server01
state     : created
size      : 3072
type      : data
id        : 4969249
kernel    : 3.12-x86_64 (hvm)
cmdline   : root=/dev/sda ro nosep console=ttyS0
datacenter: FR-SD2
vm        : server01
""")
        self.assertEqual(result.exit_code, 0)

    def test_check_size(self):
        result = disk_check_size(None, None, 2048)
        self.assertEqual(result, 2048)
        self.assertRaises(ClickException, disk_check_size, None, None, 2040)

    def test_detach(self):
        result = self.invoke_with_exceptions(disk.detach, ['data'])
        self.assertEqual(result.output.strip(),
                         "Are you sure you want to detach data? [y/N]:")
        self.assertEqual(result.exit_code, 0)

    def test_detach_forced(self):
        result = self.invoke_with_exceptions(disk.detach, ['-f', 'data'])
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
The disk is still attached to the vm 152967.
Will detach it.
Detaching your disk(s).
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_detach_background(self):
        args = ['data', '--bg', '-f']
        result = self.invoke_with_exceptions(disk.detach, args)

        self.assertEqual(result.output, """\
The disk is still attached to the vm 152967.
Will detach it.
[{'id': 200, 'step': 'WAIT'}]
""")

    def test_attach(self):
        args = ['snaptest', 'server01']
        result = self.invoke_with_exceptions(disk.attach, args)
        self.assertEqual(result.output.strip(), """\
Are you sure you want to attach disk 'snaptest' to vm 'server01'? [y/N]:\
""")
        self.assertEqual(result.exit_code, 0)

    def test_attach_forced(self):
        args = ['snaptest', 'server01', '-f']
        result = self.invoke_with_exceptions(disk.attach, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Attaching your disk(s).
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_attach_must_detach(self):
        args = ['data', 'vm1426759833']
        result = self.invoke_with_exceptions(disk.attach, args,
                                             input='y\n')
        self.assertEqual(result.output.strip(), """\
Are you sure you want to attach disk 'data' to vm 'vm1426759833'? [y/N]: y\
\nThis disk is still attached
Are you sure you want to detach data? [y/N]:""")
        self.assertEqual(result.exit_code, 0)

    def test_attach_must_detach_forced(self):
        args = ['data', 'vm1426759833', '-f']
        result = self.invoke_with_exceptions(disk.attach, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
The disk is still attached to the vm 152967.
Will detach it.
Detaching your disk.
\rProgress: [###] 100.00%  00:00:00  \
\nAttaching your disk(s).
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_attach_forced_background(self):
        args = ['snaptest', 'server01', '-f', '--bg']
        result = self.invoke_with_exceptions(disk.attach, args)
        self.assertEqual(result.output.strip(), "{'id': 200, 'step': 'WAIT'}")

        self.assertEqual(result.exit_code, 0)

    def test_update_name(self):
        args = ['data', '--name', 'data2']
        result = self.invoke_with_exceptions(disk.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your disk.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_update_kernel(self):
        args = ['data', '--kernel', '3.12-x86_64 (hvm)']
        result = self.invoke_with_exceptions(disk.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your disk.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_update_kernel_unavailable(self):
        args = ['data', '--kernel', '3.12-x86_64']
        result = self.invoke_with_exceptions(disk.update, args)

        self.assertEqual(result.output, """\
Usage: update [OPTIONS] RESOURCE

Error: Kernel 3.12-x86_64 is not available for disk data
""")

        self.assertEqual(result.exit_code, 2)

    def test_update_cmdline(self):
        args = ['data', '--cmdline',
                'root=/dev/xvda1 loglevel=4 console=hvc0 nosep ro']
        result = self.invoke_with_exceptions(disk.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your disk.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_update_snapshotprofile_ko(self):
        args = ['data', '--snapshotprofile', '7']
        result = self.invoke_with_exceptions(disk.update, args)
        self.assertEqual(result.exit_code, 2)

    def test_update_snapshotprofile(self):
        args = ['data', '--snapshotprofile', '2']
        result = self.invoke_with_exceptions(disk.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your disk.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_update_snapshotprofile_delete(self):
        args = ['data', '--delete-snapshotprofile']
        result = self.invoke_with_exceptions(disk.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your disk.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_update_snapshotprofile_conflict(self):
        args = ['data', '--delete-snapshotprofile', '--snapshotprofile', '2']
        result = self.invoke_with_exceptions(disk.update, args)
        self.assertEqual(result.exit_code, 2)

    def test_update_size(self):
        args = ['data', '--size', '5G']
        result = self.invoke_with_exceptions(disk.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your disk.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(self.api_calls['hosting.disk.update'][0][1],
                         {'size': 5120})

    def test_update_size_prefix(self):
        args = ['data', '--size', '+3G']
        result = self.invoke_with_exceptions(disk.update, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your disk.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)
        self.assertEqual(self.api_calls['hosting.disk.update'][0][1],
                         {'size': 6144})

    def test_update_background(self):
        args = ['data', '--name', 'data2', '--bg']
        result = self.invoke_with_exceptions(disk.update, args)
        self.assertEqual(result.output.strip(), "{'id': 200, 'step': 'WAIT'}")

        self.assertEqual(result.exit_code, 0)

    def test_delete(self):
        result = self.invoke_with_exceptions(disk.delete, ['data'])
        self.assertEqual(result.output.strip(),
                         "Are you sure you want to delete disk 'data'? [y/N]:")
        self.assertEqual(result.exit_code, 0)

    def test_delete_multiple(self):
        result = self.invoke_with_exceptions(disk.delete, ['data', 'snaptest'])
        self.assertEqual(result.output.strip(), """\
Are you sure you want to delete disk 'data, snaptest'? [y/N]:""")
        self.assertEqual(result.exit_code, 0)

    def test_delete_attached(self):
        result = self.invoke_with_exceptions(disk.delete, ['data', '-f'])
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
The disk is still attached to the vm 152967.
Will detach it.
Detaching your disk(s).
\rProgress: [###] 100.00%  00:00:00  \
\nDeleting your disk.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)

    def test_delete_force(self):
        result = self.invoke_with_exceptions(disk.delete, ['snaptest', '-f'])
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Deleting your disk.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)

    def test_delete_background(self):
        args = ['snaptest', '-f', '--bg']
        result = self.invoke_with_exceptions(disk.delete, args)
        self.assertEqual(result.output.strip(), """\
id        : 200
step      : WAIT""")
        self.assertEqual(result.exit_code, 0)

    def test_rollback(self):
        args = ['snaptest']
        result = self.invoke_with_exceptions(disk.rollback, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Disk rollback in progress.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)

    def test_rollback_background(self):
        args = ['snaptest', '--bg']
        result = self.invoke_with_exceptions(disk.rollback, args)
        self.assertEqual(result.output.strip(), "{'id': 200, 'step': 'WAIT'}")
        self.assertEqual(result.exit_code, 0)

    def test_migrate(self):
        args = ['snaptest']
        result = self.invoke_with_exceptions(disk.migrate, args, input='y\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to migrate disk snaptest ? [y/N]: y\
\n* Starting the migration of disk snaptest from datacenter FR-SD2 to FR-SD3
Disk migration in progress.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)

    def test_migrate_background(self):
        args = ['snaptest', '--bg', '-f']
        result = self.invoke_with_exceptions(disk.migrate, args)
        self.assertEqual(result.output.strip(), """\
* Starting the migration of disk snaptest from datacenter FR-SD2 to FR-SD3
id        : 200
step      : WAIT""")
        self.assertEqual(result.exit_code, 0)

    def test_migrate_not_available(self):
        args = ['newdisk']
        result = self.invoke_with_exceptions(disk.migrate, args, input='y\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
No datacenter is available for migration""")
        self.assertEqual(result.exit_code, 0)

    def test_migrate_noconfirm(self):
        args = ['snaptest']
        result = self.invoke_with_exceptions(disk.migrate, args, input='\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure you want to migrate disk snaptest ? [y/N]:""")
        self.assertEqual(result.exit_code, 0)

    def test_migrate_attached(self):
        args = ['data']
        result = self.invoke_with_exceptions(disk.migrate, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Cannot start the migration: disk data is attached. \
Please detach the disk before starting the migration.""")
        self.assertEqual(result.exit_code, 0)

    def test_snapshot(self):
        args = ['snaptest']
        result = self.invoke_with_exceptions(disk.snapshot, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Creating your disk.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.disk.create_from'][0][0]
        self.assertTrue(params['name'].startswith('snp'))

    def test_snapshot_background(self):
        args = ['snaptest', '--bg']
        result = self.invoke_with_exceptions(disk.snapshot, args)
        self.assertEqual(result.output.strip(), "{'id': 200, 'step': 'WAIT'}")
        self.assertEqual(result.exit_code, 0)

    def test_snapshot_name(self):
        args = ['snaptest', '--name', 'snappy']
        result = self.invoke_with_exceptions(disk.snapshot, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Creating your disk.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.disk.create_from'][0][0]
        self.assertEqual(params['name'], 'snappy')

    def test_create_default_ok(self):
        args = []
        result = self.invoke_with_exceptions(disk.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Creating your disk.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.disk.create'][0][0]
        self.assertEqual(params['type'], 'data')
        self.assertEqual(params['size'], 3072)
        self.assertTrue(params['name'].startswith('vdi'))

    def test_create_params(self):
        args = ['--name', 'newdisk', '--size', '5G', '--datacenter', 'FR-SD3',
                '--snapshotprofile', '3']
        result = self.invoke_with_exceptions(disk.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Creating your disk.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.disk.create'][0][0]
        self.assertEqual(params['datacenter_id'], 4)
        self.assertEqual(params['size'], 5120)
        self.assertEqual(params['name'], 'newdisk')
        self.assertEqual(params['snapshot_profile'], 3)

    def test_create_datacenter_closed(self):
        args = ['--name', 'newdisk', '--size', '5G', '--datacenter', 'US-BA1',
                '--snapshotprofile', '3']
        result = self.invoke_with_exceptions(disk.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Error: /!\ Datacenter US-BA1 is closed, please choose another datacenter.""")
        self.assertEqual(result.exit_code, 1)

    def test_create_datacenter_limited(self):
        args = ['--name', 'newdisk', '--size', '5G', '--datacenter', 'FR-SD2',
                '--snapshotprofile', '3']
        result = self.invoke_with_exceptions(disk.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
/!\ Datacenter FR-SD2 will be closed on 25/12/2017, please consider using \
another datacenter.
Creating your disk.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.disk.create'][0][0]
        self.assertEqual(params['datacenter_id'], 1)
        self.assertEqual(params['size'], 5120)
        self.assertEqual(params['name'], 'newdisk')
        self.assertEqual(params['snapshot_profile'], 3)

    def test_create_params_snapshot_ko(self):
        args = ['--name', 'newdisk', '--size', '5G', '--datacenter', 'FR',
                '--snapshotprofile', '7']
        result = self.invoke_with_exceptions(disk.create, args,
                                             obj=GandiContextHelper())
        self.assertEqual(result.exit_code, 2)

    def test_create_default_background(self):
        args = ['--bg']
        result = self.invoke_with_exceptions(disk.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(result.output.strip(), """\
id        : 200
step      : WAIT""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.disk.create'][0][0]
        self.assertEqual(params['type'], 'data')
        self.assertEqual(params['size'], 3072)
        self.assertTrue(params['name'].startswith('vdi'))

    def test_create_vm(self):
        args = ['--vm', 'server01']
        result = self.invoke_with_exceptions(disk.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
/!\ VM server01 datacenter will be used instead of FR-SD5.
Creating your disk.
\rProgress: [###] 100.00%  00:00:00  \
\nAttaching your disk.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['hosting.disk.create'][0][0]
        self.assertEqual(params['type'], 'data')
        self.assertEqual(params['size'], 3072)
        self.assertEqual(params['datacenter_id'], 1)
        self.assertTrue(params['name'].startswith('vdi'))

    def test_create_source(self):
        args = ['--source', 'sys_server01']
        result = self.invoke_with_exceptions(disk.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Creating your disk.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)
        params, disk_id = self.api_calls['hosting.disk.create_from'][0]
        self.assertEqual(params['type'], 'data')
        self.assertTrue(params['name'].startswith('vdi'))
        self.assertEqual(disk_id, 4969249)



""" Datacenter commands module. """

from __future__ import print_function

from gandi.cli.core.base import GandiModule
from gandi.cli.core.utils import DatacenterClosed, DatacenterLimited


class Datacenter(GandiModule):

    """ Module to handle CLI commands.

    $ gandi datacenters
    """

    @classmethod
    def list(cls, options=None):
        """List available datacenters."""
        return cls.safe_call('hosting.datacenter.list', options or {})

    @classmethod
    def list_migration_choice(cls, datacenter):
        """List available datacenters for migration from given datacenter."""
        datacenter_id = cls.usable_id(datacenter)
        dc_list = cls.list()
        available_dcs = [dc for dc in dc_list
                         if dc['id'] == datacenter_id][0]['can_migrate_to']
        choices = [dc for dc in dc_list
                   if dc['id'] in available_dcs]
        return choices

    @classmethod
    def is_opened(cls, dc_code, type_):
        """List opened datacenters for given type."""
        options = {'dc_code': dc_code, '%s_opened' % type_: True}
        datacenters = cls.safe_call('hosting.datacenter.list', options)
        if not datacenters:
            # try with ISO code
            options = {'iso': dc_code, '%s_opened' % type_: True}
            datacenters = cls.safe_call('hosting.datacenter.list', options)
            if not datacenters:
                raise DatacenterClosed(r'/!\ Datacenter %s is closed, please '
                                       'choose another datacenter.' % dc_code)

        datacenter = datacenters[0]
        if datacenter.get('%s_closed_for' % type_) == 'NEW':
            dc_close_date = datacenter.get('deactivate_at', '')
            if dc_close_date:
                dc_close_date = dc_close_date.strftime('%d/%m/%Y')
            raise DatacenterLimited(dc_close_date)

    @classmethod
    def filtered_list(cls, name=None, obj=None):
        """List datacenters matching name and compatible
        with obj"""
        options = {}
        if name:
            options['id'] = cls.usable_id(name)

        def obj_ok(dc, obj):
            if not obj or obj['datacenter_id'] == dc['id']:
                return True
            return False

        return [x for x in cls.list(options) if obj_ok(x, obj)]

    @classmethod
    def from_iso(cls, iso):
        """Retrieve the first datacenter id associated to an ISO."""
        result = cls.list({'sort_by': 'id ASC'})
        dc_isos = {}
        for dc in result:
            if dc['iso'] not in dc_isos:
                dc_isos[dc['iso']] = dc['id']

        return dc_isos.get(iso)

    @classmethod
    def from_name(cls, name):
        """Retrieve datacenter id associated to a name."""
        result = cls.list()
        dc_names = {}
        for dc in result:
            dc_names[dc['name']] = dc['id']

        return dc_names.get(name)

    @classmethod
    def from_country(cls, country):
        """Retrieve the first datacenter id associated to a country."""
        result = cls.list({'sort_by': 'id ASC'})
        dc_countries = {}
        for dc in result:
            if dc['country'] not in dc_countries:
                dc_countries[dc['country']] = dc['id']

        return dc_countries.get(country)

    @classmethod
    def from_dc_code(cls, dc_code):
        """Retrieve the datacenter id associated to a dc_code"""
        result = cls.list()
        dc_codes = {}
        for dc in result:
            if dc.get('dc_code'):
                dc_codes[dc['dc_code']] = dc['id']

        return dc_codes.get(dc_code)

    @classmethod
    def usable_id(cls, id):
        """ Retrieve id from input which can be ISO, name, country, dc_code."""
        try:
            # id is maybe a dc_code
            qry_id = cls.from_dc_code(id)
            if not qry_id:
                # id is maybe a ISO
                qry_id = cls.from_iso(id)
                if qry_id:
                    cls.deprecated('ISO code for datacenter filter use '
                                   'dc_code instead')
            if not qry_id:
                # id is maybe a country
                qry_id = cls.from_country(id)
            if not qry_id:
                qry_id = int(id)
        except Exception:
            qry_id = None

        if not qry_id:
            msg = 'unknown identifier %s' % id
            cls.error(msg)

        return qry_id



# -*- coding: utf-8 -*-
""" Configuration handler class declaration. """

import os
import sys
import yaml
import os.path
from distutils.dir_util import mkpath

try:
    from yaml import CSafeLoader as YAMLLoader
except ImportError:
    from yaml import SafeLoader as YAMLLoader

import click


class GandiConfig(object):

    """ Base class for yaml configuration.

    Manage
    - read/write configuration files/values
    - handle two scopes : global/local

    """

    _conffiles = {}
    home_config = os.environ.get('GANDI_CONFIG',
                                 '~/.config/gandi/config.yaml')
    local_config = '.gandi.config.yaml'

    apienvs = {
        'ote': 'https://rpc.ote.gandi.net/xmlrpc/',
        'production': 'https://rpc.gandi.net/xmlrpc/',
    }
    default_apienv = 'production'

    @classmethod
    def load_config(cls):
        """ Load global and local configuration files and update if needed."""
        config_file = os.path.expanduser(cls.home_config)
        global_conf = cls.load(config_file, 'global')
        cls.load(cls.local_config, 'local')
        # update global configuration if needed
        cls.update_config(config_file, global_conf)

    @classmethod
    def update_config(cls, config_file, config):
        """ Update configuration if needed. """
        need_save = False
        # delete old env key
        if 'api' in config and 'env' in config['api']:
            del config['api']['env']
            need_save = True
        # convert old ssh_key configuration entry
        ssh_key = config.get('ssh_key')
        sshkeys = config.get('sshkey')
        if ssh_key and not sshkeys:
            config.update({'sshkey': [ssh_key]})
            need_save = True
        elif ssh_key and sshkeys:
            config.update({'sshkey': sshkeys.append(ssh_key)})
            need_save = True
        # remove old value
        if ssh_key:
            del config['ssh_key']
            need_save = True

        # save to disk
        if need_save:
            cls.save(config_file, config)

    @classmethod
    def load(cls, filename, name=None):
        """ Load yaml configuration from filename. """
        if not os.path.exists(filename):
            return {}
        name = name or filename

        if name not in cls._conffiles:
            with open(filename) as fdesc:
                content = yaml.load(fdesc, YAMLLoader)
                # in case the file is empty
                if content is None:
                    content = {}
                cls._conffiles[name] = content

        return cls._conffiles[name]

    @classmethod
    def save(cls, filename, config):
        """ Save configuration to yaml file. """
        mode = os.O_WRONLY | os.O_TRUNC | os.O_CREAT
        with os.fdopen(os.open(filename, mode, 0o600), 'w') as fname:
            yaml.safe_dump(config, fname, indent=4, default_flow_style=False)

    @classmethod
    def _del(cls, scope, key, separator='.', conf=None):
        orig_key = key

        key = key.split(separator)
        if not conf:
            conf = cls._conffiles.get(scope, {})

        if separator not in orig_key:
            if orig_key in conf:
                del conf[orig_key]
                return

        for k in key:
            if k not in conf:
                return
            else:
                cls._del(scope, separator.join([k1 for k1 in key if k1 != k]),
                         conf=conf[k])
                return

    @classmethod
    def delete(cls, global_, key):
        """ Delete key/value pair from configuration file. """
        # first retrieve current configuration
        scope = 'global' if global_ else 'local'
        config = cls._conffiles.get(scope, {})
        cls._del(scope, key)
        conf_file = cls.home_config if global_ else cls.local_config
        # save configuration to file
        cls.save(os.path.expanduser(conf_file), config)

    @classmethod
    def _set(cls, scope, key, val, separator='.'):
        orig_key = key

        key = key.split(separator)
        value = cls._conffiles.get(scope, {})
        if separator not in orig_key:
            value[orig_key] = val
            return

        for k in key:
            if k not in value:
                value[k] = {}
                last_val = value
                value = value[k]
            else:
                last_val = value
                value = value[k]
        last_val[k] = val

    @classmethod
    def _get(cls, scope, key, default=None, separator='.'):
        key = key.split(separator)
        value = cls._conffiles.get(scope, {})
        if not value:
            return default

        try:
            for k in key:
                value = value[k]
            return value
        except KeyError:
            return default

    @classmethod
    def get(cls, key, default=None, separator='.', global_=False):
        """ Retrieve a key value from loaded configuration.

        Order of search if global_=False:
        1/ environnment variables
        2/ local configuration
        3/ global configuration
        """
        # first check environnment variables
        # if we're not in global scope
        if not global_:
            ret = os.environ.get(key.upper().replace('.', '_'))
            if ret is not None:
                return ret

        # then check in local and global configuration unless global_=True
        scopes = ['global'] if global_ else ['local', 'global']
        for scope in scopes:
            ret = cls._get(scope, key, default, separator)
            if ret is not None and ret != default:
                return ret

        if ret is None or ret == default:
            return default

    @classmethod
    def configure(cls, global_, key, val):
        """ Update and save configuration value to file. """
        # first retrieve current configuration
        scope = 'global' if global_ else 'local'
        if scope not in cls._conffiles:
            cls._conffiles[scope] = {}
        config = cls._conffiles.get(scope, {})
        # apply modification to fields
        cls._set(scope, key, val)
        conf_file = cls.home_config if global_ else cls.local_config
        # save configuration to file
        cls.save(os.path.expanduser(conf_file), config)

    @classmethod
    def list(cls, global_):
        """ Return configuration file content. """
        scope = 'global' if global_ else 'local'
        return cls._conffiles.get(scope, {})

    @classmethod
    def init_config(cls):
        """ Initialize Gandi CLI configuration.

        Create global configuration directory with API credentials

        """
        try:
            # first load current conf and only overwrite needed params
            # we don't want to reset everything
            config_file = os.path.expanduser(cls.home_config)
            config = cls.load(config_file, 'global')
            cls._del('global', 'api.env')

            hidden_apikey = '%s...' % cls.get('api.key', '')[:6]
            apikey = click.prompt('Api key (xmlrpc)',
                                  default=hidden_apikey)
            if apikey == hidden_apikey:
                # if default value then use actual value not hidden one
                apikey = cls.get('api.key')
            env_choice = click.Choice(list(cls.apienvs.keys()))
            apienv = click.prompt('Environnment [production]/ote',
                                  default=cls.default_apienv,
                                  type=env_choice,
                                  show_default=False)
            sshkey = click.prompt('SSH keyfile',
                                  default='~/.ssh/id_rsa.pub')
            hidden_apikeyrest = '%s...' % cls.get('apirest.key', '')[:6]
            apikeyrest = click.prompt('Api key (REST)',
                                      default=hidden_apikeyrest)
            if apikeyrest == hidden_apikeyrest:
                # if default value then use actual value not hidden one
                apikeyrest = cls.get('apirest.key')

            config.update({
                'api': {'key': apikey,
                        'host': cls.apienvs[apienv]},
            })
            if apikeyrest:
                config.update({
                    'apirest': {'key': apikeyrest},
                })

            if sshkey is not None:
                sshkey_file = os.path.expanduser(sshkey)
                if os.path.exists(sshkey_file):
                    config['sshkey'] = [sshkey_file]

            directory = os.path.expanduser(os.path.dirname(config_file))
            if not os.path.exists(directory):
                mkpath(directory, 0o700)

            # save to disk
            cls.save(config_file, config)
            # load in memory
            cls.load(config_file, 'global')
        except (KeyboardInterrupt, click.exceptions.Abort):
            cls.echo('Aborted.')
            sys.exit(1)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from functools import partial

import re

from ..compat import mock, StringIO, ConfigParser
from .base import CommandTestCase
from gandi.cli.commands import paas
from gandi.cli.core.base import GandiContextHelper


def _mock_output(git_content, command, *args, **kwargs):
    buf = StringIO(git_content)
    config = ConfigParser.ConfigParser()
    try:
        config.read_file(buf)
    except Exception:
        config.readfp(buf)
    if command == 'git config --local --get remote.gandi.url':
        try:
            return config.get('remote "gandi"', 'url')
        except Exception:
            return ''

    if command == 'git config --local --get remote.origin.url':
        try:
            return config.get('remote "origin"', 'url')
        except Exception:
            return ''

    if command == 'git config --local --get remote.$(git config --local --get branch.stable.remote).url':  # noqa
        try:
            return config.get('remote "production"', 'url')
        except Exception:
            return ''

    if command == 'git config --local --get remote.$(git config --local --get branch.master.remote).url':  # noqa
        try:
            return config.get('remote "origin"', 'url')
        except Exception:
            return ''

    return ''


class PaasTestCase(CommandTestCase):

    def test_list(self):
        result = self.invoke_with_exceptions(paas.list, [])

        self.assertEqual(result.output, """\
name       : paas_owncloud
state      : halted
vhost      : aa3e0e26f8.url-de-test.ws
vhost      : cloud.cat.lol
----------
name       : paas_cozycloud
state      : running
vhost      : 187832c2b34.testurl.ws
vhost      : cloud.iheartcli.com
vhost      : cli.sexy
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_id(self):
        result = self.invoke_with_exceptions(paas.list, ['--id'])

        self.assertEqual(result.output, """\
name       : paas_owncloud
state      : halted
id         : 126276
vhost      : aa3e0e26f8.url-de-test.ws
vhost      : cloud.cat.lol
----------
name       : paas_cozycloud
state      : running
id         : 163744
vhost      : 187832c2b34.testurl.ws
vhost      : cloud.iheartcli.com
vhost      : cli.sexy
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_type(self):
        result = self.invoke_with_exceptions(paas.list, ['--type'])

        self.assertEqual(result.output, """\
name       : paas_owncloud
state      : halted
type       : phpmysql
vhost      : aa3e0e26f8.url-de-test.ws
vhost      : cloud.cat.lol
----------
name       : paas_cozycloud
state      : running
type       : nodejsmongodb
vhost      : 187832c2b34.testurl.ws
vhost      : cloud.iheartcli.com
vhost      : cli.sexy
""")

        self.assertEqual(result.exit_code, 0)

    def test_list_filter_state(self):
        result = self.invoke_with_exceptions(paas.list, ['--state', 'halted'])

        self.assertEqual(result.output, """\
name       : paas_owncloud
state      : halted
vhost      : aa3e0e26f8.url-de-test.ws
vhost      : cloud.cat.lol
""")

        self.assertEqual(result.exit_code, 0)

    def test_info(self):
        result = self.invoke_with_exceptions(paas.info, ['paas_cozycloud'])

        self.assertEqual(result.output, """\
name       : paas_cozycloud
type       : nodejsmongodb
size       : s
console    : 185290@console.dc2.gpaas.net
git_server : git.dc2.gpaas.net
sftp_server: sftp.dc2.gpaas.net
vhost      : 187832c2b34.testurl.ws
vhost      : cloud.iheartcli.com
vhost      : cli.sexy
datacenter : LU-BI1
quota used : 0.5%
snapshot   :
""")

        self.assertEqual(result.exit_code, 0)

    def test_info_stat(self):
        args = ['paas_cozycloud', '--stat']
        result = self.invoke_with_exceptions(paas.info, args)

        self.assertEqual(result.output, """\
name       : paas_cozycloud
type       : nodejsmongodb
size       : s
console    : 185290@console.dc2.gpaas.net
git_server : git.dc2.gpaas.net
sftp_server: sftp.dc2.gpaas.net
vhost      : 187832c2b34.testurl.ws
vhost      : cloud.iheartcli.com
vhost      : cli.sexy
datacenter : LU-BI1
quota used : 0.5%
snapshot   :
cache      :
\thit  : 0.0%
\tmiss : 100.0%
\tnot  : 0.0%
\tpass : 0.0%
""")

        self.assertEqual(result.exit_code, 0)

    def test_types(self):
        result = self.invoke_with_exceptions(paas.types, [])

        self.assertEqual(result.output, """\
phpmysql
phppgsql
nodejspgsql
nodejsmongodb
nodejsmysql
phpmongodb
pythonmysql
pythonpgsql
pythonmongodb
rubymysql
rubypgsql
rubymongodb
""")

        self.assertEqual(result.exit_code, 0)

    def test_console(self):
        result = self.invoke_with_exceptions(paas.console, ['paas_cozycloud'])

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
/!\ Use ~. ssh escape key to exit.
Activation of the console on your PaaS instance
\rProgress: [###] 100.00%  00:00:00  \n\
ssh 185290@console.dc2.gpaas.net""")

        self.assertEqual(result.exit_code, 0)

    def test_clone_missing(self):
        result = self.invoke_with_exceptions(paas.clone, [])

        self.assertEqual(result.output, """\
Usage: clone [OPTIONS] NAME
Try "clone --help" for help.

Error: Missing argument "NAME".
""")

        self.assertEqual(result.exit_code, 2)

    def test_clone(self):
        with mock.patch('gandi.cli.modules.vhost.os.chdir',
                        create=True) as mock_chdir:
            mock_chdir.return_value = mock.MagicMock()

            result = self.invoke_with_exceptions(paas.clone, ['cli.sexy'])

        self.assertEqual(result.output, """\
git clone ssh+git://185290@git.dc2.gpaas.net/default.git cli.sexy \
--origin gandi
Use `git push gandi master` to push your code to the instance.
Then `$ gandi deploy` to build and deploy your application.
""")

        self.assertEqual(result.exit_code, 0)

    def test_clone_directory(self):
        with mock.patch('gandi.cli.modules.vhost.os.chdir',
                        create=True) as mock_chdir:
            mock_chdir.return_value = mock.MagicMock()

            args = ['cli.sexy', '--directory', 'project']
            result = self.invoke_with_exceptions(paas.clone, args)

        self.assertEqual(result.output, """\
git clone ssh+git://185290@git.dc2.gpaas.net/default.git project --origin gandi
Use `git push gandi master` to push your code to the instance.
Then `$ gandi deploy` to build and deploy your application.
""")

        self.assertEqual(result.exit_code, 0)

    def test_clone_vhost(self):
        with mock.patch('gandi.cli.modules.vhost.os.chdir',
                        create=True) as mock_chdir:
            mock_chdir.return_value = mock.MagicMock()

            args = ['paas_cozycloud', '--vhost', 'cli.sexy']
            result = self.invoke_with_exceptions(paas.clone, args)

        self.assertEqual(result.output, """\
git clone ssh+git://185290@git.dc2.gpaas.net/cli.sexy.git cli.sexy \
--origin gandi
Use `git push gandi master` to push your code to the instance.
Then `$ gandi deploy` to build and deploy your application.
""")

        self.assertEqual(result.exit_code, 0)

    def test_attach(self):
        result = self.invoke_with_exceptions(paas.attach, ['paas_cozycloud'])

        self.assertEqual(result.output, """\
git remote add gandi ssh+git://185290@git.dc2.gpaas.net/default.git
Added remote `gandi` to your local git repository.
Use `git push gandi master` to push your code to the instance.
Then `$ gandi deploy` to build and deploy your application.
""")

        self.assertEqual(result.exit_code, 0)

    def test_attach_remote(self):
        args = ['paas_cozycloud', '--remote', 'production']
        result = self.invoke_with_exceptions(paas.attach, args)

        self.assertEqual(result.output, """\
git remote add production ssh+git://185290@git.dc2.gpaas.net/default.git
Added remote `production` to your local git repository.
Use `git push production master` to push your code to the instance.
Then `$ gandi deploy` to build and deploy your application.
""")

        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.base.GandiModule.exec_output')
    def test_deploy_invalid_remote_empty(self, mock_exec_output):
        args = []

        git_content = """
[blabla]
dummy=dududududud
"""
        mock_exec_output.side_effect = partial(_mock_output, git_content)
        result = self.invoke_with_exceptions(paas.deploy, args)

        self.assertEqual(result.output, """\
Error: Could not find git remote to extract deploy url from.
This usually happens when:
- the current directory has no Simple Hosting git remote attached,
  in this case, please see $ gandi paas attach --help
- the local branch being deployed hasn't been pushed \
to the remote repository yet,
  in this case, please try $ git push <remote> master

Otherwise, it's recommended to use the --remote and/or --branch options:
$ gandi deploy --remote <remote> [--branch <branch>]
""")

        self.assertEqual(result.exit_code, 2)

    @mock.patch('gandi.cli.core.base.GandiModule.exec_output')
    def test_deploy_invalid_remote_content(self, mock_exec_output):
        args = []
        self.maxDiff = None

        git_content = """
[remote "origin"]
fetch = +refs/heads/*:refs/remotes/origin/*
url = https://github.com/Gandi/gandi.cli.git
[branch "master"]
remote = origin
merge = refs/heads/master
"""
        mock_exec_output.side_effect = partial(_mock_output, git_content)
        result = self.invoke_with_exceptions(paas.deploy, args)

        self.assertEqual(result.output, """\
Error: https://github.com/Gandi/gandi.cli.git \
is not a valid Simple Hosting git remote.
This usually happens when:
- the current directory has no Simple Hosting git remote attached,
  in this case, please see $ gandi paas attach --help
- the local branch being deployed hasn't been pushed \
to the remote repository yet,
  in this case, please try $ git push <remote> master

Otherwise, it's recommended to use the --remote and/or --branch options:
$ gandi deploy --remote <remote> [--branch <branch>]
""")

        self.assertEqual(result.exit_code, 2)

    @mock.patch('gandi.cli.core.base.GandiModule.exec_output')
    def test_deploy(self, mock_exec_output):
        args = []

        git_content = """
[remote "gandi"]
fetch = +refs/heads/*:refs/remotes/gandi/*
url = ssh+git://185290@git.dc2.gpaas.net/default.git
"""
        mock_exec_output.side_effect = partial(_mock_output, git_content)
        result = self.invoke_with_exceptions(paas.deploy, args)

        self.assertEqual(result.output, """\
ssh 185290@git.dc2.gpaas.net 'deploy default.git master'
""")

        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.base.GandiModule.exec_output')
    def test_deploy_remote(self, mock_exec_output):
        args = ['--remote', 'origin']

        git_content = """
[remote "origin"]
fetch = +refs/heads/*:refs/remotes/origin/*
url = ssh+git://185290@git.dc2.gpaas.net/default.git
[branch "master"]
remote = origin
merge = refs/heads/master
"""
        mock_exec_output.side_effect = partial(_mock_output, git_content)
        result = self.invoke_with_exceptions(paas.deploy, args)

        self.assertEqual(result.output, """\
ssh 185290@git.dc2.gpaas.net 'deploy default.git master'
""")

        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.base.GandiModule.exec_output')
    def test_deploy_guess_remote_with_branch(self, mock_exec_output):
        args = ['--branch', 'stable']

        git_content = """
[remote "origin"]
fetch = +refs/heads/*:refs/remotes/origin/*
url = https://github.com/Gandi/gandi.cli.git
[remote "production"]
fetch = +refs/heads/*:refs/remotes/production/*
url = ssh+git://185290@git.dc2.gpaas.net/default.git
[branch "master"]
remote = origin
merge = refs/heads/master
[branch "stable"]
remote = production
merge = refs/heads/stable
"""
        mock_exec_output.side_effect = partial(_mock_output, git_content)
        result = self.invoke_with_exceptions(paas.deploy, args)

        self.assertEqual(result.output, """\
ssh 185290@git.dc2.gpaas.net 'deploy default.git stable'
""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_unknown(self):
        result = self.invoke_with_exceptions(paas.delete, ['unknown_paas'])

        self.assertEqual(result.output, """\
Sorry PaaS instance unknown_paas does not exist
Please use one of the following: ['paas_owncloud', 'paas_cozycloud', \
'126276', '163744', 'aa3e0e26f8.url-de-test.ws', 'cloud.cat.lol', \
'187832c2b34.testurl.ws', 'cloud.iheartcli.com', 'cli.sexy']
""")

        self.assertEqual(result.exit_code, 0)

    def test_delete(self):
        result = self.invoke_with_exceptions(paas.delete, ['paas_owncloud'])
        self.assertEqual(result.output.strip(), """\
Are you sure to delete PaaS instance 'paas_owncloud'? [y/N]:""")
        self.assertEqual(result.exit_code, 0)

    def test_delete_ko(self):
        args = ['paas_owncloud']
        result = self.invoke_with_exceptions(paas.delete, args, input='N\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure to delete PaaS instance 'paas_owncloud'? [y/N]: N\
""")
        self.assertEqual(result.exit_code, 0)

    def test_delete_ok(self):
        args = ['paas_owncloud']
        result = self.invoke_with_exceptions(paas.delete, args, input='y\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure to delete PaaS instance 'paas_owncloud'? [y/N]: y
Deleting your PaaS instance.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)

    def test_delete_background(self):
        args = ['paas_owncloud', '--force', '--background']
        result = self.invoke_with_exceptions(paas.delete, args)
        self.assertEqual(result.output, """\
id        : 200
step      : WAIT
""")
        self.assertEqual(result.exit_code, 0)

    def test_create_default(self):
        args = []
        result = self.invoke_with_exceptions(paas.create, args,
                                             obj=GandiContextHelper(),
                                             input='ploki\nploki\n')

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'paas\d+', 'paas', output), """\
password: \nRepeat for confirmation: \n\
Creating your PaaS instance.
\rProgress: [###] 100.00%  00:00:00  \n\
Your PaaS instance paas has been created.""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['paas.create'][0][0]
        self.assertEqual(params['datacenter_id'], 3)
        self.assertEqual(params['size'], 's')
        self.assertEqual(params['duration'], '1m')
        self.assertEqual(params['password'], 'ploki')
        self.assertTrue(params['name'].startswith('paas'))

    def test_create_size(self):
        args = ['--size', 's+']
        result = self.invoke_with_exceptions(paas.create, args,
                                             obj=GandiContextHelper(),
                                             input='ploki\nploki\n')

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'paas\d+', 'paas', output), """\
password: \nRepeat for confirmation: \n\
Creating your PaaS instance.
\rProgress: [###] 100.00%  00:00:00  \n\
Your PaaS instance paas has been created.""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['paas.create'][0][0]
        self.assertEqual(params['datacenter_id'], 3)
        self.assertEqual(params['size'], 's+')
        self.assertEqual(params['duration'], '1m')
        self.assertEqual(params['password'], 'ploki')
        self.assertTrue(params['name'].startswith('paas'))

    def test_create_name(self):
        args = ['--name', '123456']
        result = self.invoke_with_exceptions(paas.create, args,
                                             obj=GandiContextHelper(),
                                             input='ploki\nploki\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
password: \nRepeat for confirmation: \n\
Creating your PaaS instance.
\rProgress: [###] 100.00%  00:00:00  \n\
Your PaaS instance 123456 has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_name_vhost(self):
        self.maxDiff = None
        args = ['--name', '123456', '--vhosts', 'ploki.fr', '--ssl']
        with mock.patch('gandi.cli.modules.vhost.os.chdir',
                        create=True) as mock_chdir:
            mock_chdir.return_value = mock.MagicMock()

            result = self.invoke_with_exceptions(paas.create, args,
                                                 obj=GandiContextHelper(),
                                                 input='ploki\nploki\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
password: \nRepeat for confirmation: \n\
There is no certificate for ploki.fr.
Create the certificate with (for exemple) :
$ gandi certificate create --cn ploki.fr --type std \n\
Or relaunch the current command with --poll-cert option
Creating your PaaS instance.
\rProgress: [###] 100.00%  00:00:00  \n\
Your PaaS instance 123456 has been created.
Creating a new vhost.
\rProgress: [###] 100.00%  00:00:00  \n\
Your vhost ploki.fr has been created.""")

        self.assertEqual(result.exit_code, 0)

    def test_create_name_vhost_ssl(self):
        self.maxDiff = None
        args = ['--name', '123456', '--vhosts', 'inter.net', '--ssl']
        with mock.patch('gandi.cli.modules.vhost.os.chdir',
                        create=True) as mock_chdir:
            mock_chdir.return_value = mock.MagicMock()

            result = self.invoke_with_exceptions(paas.create, args,
                                                 obj=GandiContextHelper(),
                                                 input='ploki\nploki\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
password: \nRepeat for confirmation: \n\
Please give the private key for certificate id 706 (CN: inter.net)""")

        self.assertEqual(result.exit_code, 0)

    def test_create_datacenter_limited(self):
        args = ['--datacenter', 'FR-SD2']
        result = self.invoke_with_exceptions(paas.create, args,
                                             obj=GandiContextHelper(),
                                             input='ploki\nploki\n')

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'paas\d+', 'paas', output), """\
/!\ Datacenter FR-SD2 will be closed on 25/12/2017, please consider using \
another datacenter.
password: \nRepeat for confirmation: \n\
Creating your PaaS instance.
\rProgress: [###] 100.00%  00:00:00  \n\
Your PaaS instance paas has been created.""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['paas.create'][0][0]
        self.assertEqual(params['datacenter_id'], 1)
        self.assertEqual(params['size'], 's')
        self.assertEqual(params['duration'], '1m')
        self.assertEqual(params['password'], 'ploki')
        self.assertTrue(params['name'].startswith('paas'))

        self.assertEqual(result.exit_code, 0)

    def test_create_datacenter_closed(self):
        args = ['--datacenter', 'US-BA1']
        result = self.invoke_with_exceptions(paas.create, args,
                                             obj=GandiContextHelper(),
                                             input='ploki\nploki\n')

        output = re.sub(r'\[#+\]', '[###]', result.output.strip())

        self.assertEqual(re.sub(r'paas\d+', 'paas', output), """\
Error: /!\ Datacenter US-BA1 is closed, please choose another datacenter.""")

        self.assertEqual(result.exit_code, 1)

    def test_update(self):
        args = ['paas_owncloud']
        result = self.invoke_with_exceptions(paas.update, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your PaaS instance.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_update_for_upgrade(self):
        args = ['paas_owncloud', '--upgrade']
        result = self.invoke_with_exceptions(paas.update, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your PaaS instance.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['paas.update'][0][1]
        self.assertEqual(params['upgrade'], True)

    def test_update_snapshotprofile_conflict(self):
        args = ['paas_owncloud', '--delete-snapshotprofile',
                '--snapshotprofile', '7']
        result = self.invoke_with_exceptions(paas.update, args)
        self.assertEqual(result.exit_code, 2)

    def test_update_password(self):
        args = ['paas_owncloud', '--password']
        result = self.invoke_with_exceptions(paas.update, args,
                                             obj=GandiContextHelper(),
                                             input='ploki\nploki\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
password: \nRepeat for confirmation: \n\
Updating your PaaS instance.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_update_delete_snapshotprofile(self):
        args = ['paas_owncloud', '--delete-snapshotprofile']
        result = self.invoke_with_exceptions(paas.update, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Updating your PaaS instance.
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['paas.update'][0][1]
        self.assertFalse('upgrade' in params)

    def test_update_background(self):
        args = ['paas_owncloud', '--bg']
        result = self.invoke_with_exceptions(paas.update, args)

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
{'id': 200, 'step': 'WAIT'}""")

        self.assertEqual(result.exit_code, 0)

    def test_restart_unknown(self):
        args = ['unknown_paas']
        result = self.invoke_with_exceptions(paas.restart, args)
        self.assertEqual(result.output, """\
Sorry PaaS instance unknown_paas does not exist
Please use one of the following: ['paas_owncloud', 'paas_cozycloud', \
'126276', '163744', 'aa3e0e26f8.url-de-test.ws', 'cloud.cat.lol', \
'187832c2b34.testurl.ws', 'cloud.iheartcli.com', 'cli.sexy']
""")

    def test_restart_prompt_ko(self):
        args = ['paas_owncloud']
        result = self.invoke_with_exceptions(paas.restart, args, input='N\n')
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure to restart PaaS instance 'paas_owncloud'? [y/N]: N\
""")
        self.assertEqual(result.exit_code, 0)

    def test_restart(self):
        args = ['paas_owncloud', '--force']
        result = self.invoke_with_exceptions(paas.restart, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Restarting your PaaS instance.
\rProgress: [###] 100.00%  00:00:00""")
        self.assertEqual(result.exit_code, 0)

    def test_restart_background(self):
        args = ['paas_owncloud', '--force', '--bg']
        result = self.invoke_with_exceptions(paas.restart, args)
        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
id        : 200
step      : WAIT""")
        self.assertEqual(result.exit_code, 0)



import re

from .base import CommandTestCase
from gandi.cli.commands import mail
from gandi.cli.core.base import GandiContextHelper


class MailTestCase(CommandTestCase):

    def test_list(self):

        result = self.invoke_with_exceptions(mail.list, ['iheartcli.com'])

        self.assertEqual(result.output, """admin
""")

        self.assertEqual(result.exit_code, 0)

    def test_info(self):
        args = ['admin@iheartcli.com']
        result = self.invoke_with_exceptions(mail.info, args)

        self.assertEqual(result.output, """login           : admin
aliases         :
fallback email  :
quota usage     : 233 KiB / unlimited
responder active: no
responder text  :
""")

        self.assertEqual(result.exit_code, 0)

    def test_create(self):
        args = ['contact@iheartcli.com', '--quota', '2', '--fallback',
                'admin@cli.sexy', '--alias', 'god@iheartcli.com']
        result = self.invoke_with_exceptions(mail.create, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')

        self.assertEqual(result.output, """password: \
\nRepeat for confirmation: \
\nCreating your mailbox.
Creating aliases.
""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['domain.mailbox.create'][0][2]
        self.assertEqual(params['password'], 'plokiploki')
        self.assertEqual(params['quota'], 2)
        self.assertEqual(params['fallback_email'], 'admin@cli.sexy')

    def test_create_with_password(self):
        args = ['contact2@iheartcli.com', '--quota', '2', '--fallback',
                'admin@cli.sexy', '--alias', 'god@iheartcli.com',
                '--password', 'password_for_create']
        result = self.invoke_with_exceptions(mail.create, args,
                                             obj=GandiContextHelper())

        self.assertEqual(result.output, """Creating your mailbox.
Creating aliases.
""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['domain.mailbox.create'][0][2]
        self.assertEqual(params['password'], 'password_for_create')
        self.assertEqual(params['quota'], 2)
        self.assertEqual(params['fallback_email'], 'admin@cli.sexy')

    def test_delete_force(self):
        args = ['admin@iheartcli.com', '--force']
        result = self.invoke_with_exceptions(mail.delete, args)

        self.assertEqual(result.output, '')

        self.assertEqual(result.exit_code, 0)

    def test_delete(self):
        args = ['admin@iheartcli.com']
        result = self.invoke_with_exceptions(mail.delete, args, input='y\n')

        self.assertEqual(result.output, """\
Are you sure to delete the mailbox admin@iheartcli.com ? [y/N]: y
""")

        self.assertEqual(result.exit_code, 0)

    def test_delete_force_refused(self):
        args = ['admin@iheartcli.com']
        result = self.invoke_with_exceptions(mail.delete, args, input='\n')

        self.assertEqual(result.output, """\
Are you sure to delete the mailbox admin@iheartcli.com ? [y/N]: \n""")

        self.assertEqual(result.exit_code, 0)

    def test_udpate(self):
        args = ['admin@iheartcli.com', '--quota', '2', '--fallback',
                'admin@cli.sexy', '--alias-add', 'doge@iheartcli.com',
                '--alias-del', 'god@iheartcli.com', '--password']

        result = self.invoke_with_exceptions(mail.update, args,
                                             obj=GandiContextHelper(),
                                             input='plokiploki\nplokiploki\n')

        self.assertEqual(result.output, """password: \
\nRepeat for confirmation: \
\nUpdating your mailbox.
Updating aliases.
""")

        self.assertEqual(result.exit_code, 0)
        params = self.api_calls['domain.mailbox.update'][0][2]
        self.assertEqual(params['password'], 'plokiploki')
        self.assertEqual(params['quota'], 2)
        self.assertEqual(params['fallback_email'], 'admin@cli.sexy')

    def test_purge(self):
        args = ['admin@iheartcli.com']
        result = self.invoke_with_exceptions(mail.purge, args, input='y\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure to purge mailbox admin@iheartcli.com ? [y/N]: y
Purging in progress
\rProgress: [###] 100.00%  00:00:00""")

        self.assertEqual(result.exit_code, 0)

    def test_purge_alias(self):
        args = ['admin@iheartcli.com', '--alias']
        result = self.invoke_with_exceptions(mail.purge, args, input='y\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure to purge all aliases for mailbox admin@iheartcli.com ? [y/N]: y\
""")

        self.assertEqual(result.exit_code, 0)

    def test_purge_alias_refused(self):
        args = ['admin@iheartcli.com', '--alias']
        result = self.invoke_with_exceptions(mail.purge, args, input='\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure to purge all aliases for mailbox admin@iheartcli.com ? [y/N]:\
""")

        self.assertEqual(result.exit_code, 0)

    def test_purge_refused(self):
        args = ['admin@iheartcli.com']
        result = self.invoke_with_exceptions(mail.purge, args, input='\n')

        self.assertEqual(re.sub(r'\[#+\]', '[###]',
                                result.output.strip()), """\
Are you sure to purge mailbox admin@iheartcli.com ? [y/N]:\
""")

        self.assertEqual(result.exit_code, 0)



""" Certificate commands module. """

import os
import re
from click import UsageError
from gandi.cli.core.base import GandiModule


CROSSED_PEM = '''-----BEGIN CERTIFICATE-----
MIIFdzCCBF+gAwIBAgIQE+oocFv07O0MNmMJgGFDNjANBgkqhkiG9w0BAQwFADBv
MQswCQYDVQQGEwJTRTEUMBIGA1UEChMLQWRkVHJ1c3QgQUIxJjAkBgNVBAsTHUFk
ZFRydXN0IEV4dGVybmFsIFRUUCBOZXR3b3JrMSIwIAYDVQQDExlBZGRUcnVzdCBF
eHRlcm5hbCBDQSBSb290MB4XDTAwMDUzMDEwNDgzOFoXDTIwMDUzMDEwNDgzOFow
gYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpOZXcgSmVyc2V5MRQwEgYDVQQHEwtK
ZXJzZXkgQ2l0eTEeMBwGA1UEChMVVGhlIFVTRVJUUlVTVCBOZXR3b3JrMS4wLAYD
VQQDEyVVU0VSVHJ1c3QgUlNBIENlcnRpZmljYXRpb24gQXV0aG9yaXR5MIICIjAN
BgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEAgBJlFzYOw9sIs9CsVw127c0n00yt
UINh4qogTQktZAnczomfzD2p7PbPwdzx07HWezcoEStH2jnGvDoZtF+mvX2do2NC
tnbyqTsrkfjib9DsFiCQCT7i6HTJGLSR1GJk23+jBvGIGGqQIjy8/hPwhxR79uQf
jtTkUcYRZ0YIUcuGFFQ/vDP+fmyc/xadGL1RjjWmp2bIcmfbIWax1Jt4A8BQOujM
8Ny8nkz+rwWWNR9XWrf/zvk9tyy29lTdyOcSOk2uTIq3XJq0tyA9yn8iNK5+O2hm
AUTnAU5GU5szYPeUvlM3kHND8zLDU+/bqv50TmnHa4xgk97Exwzf4TKuzJM7UXiV
Z4vuPVb+DNBpDxsP8yUmazNt925H+nND5X4OpWaxKXwyhGNVicQNwZNUMBkTrNN9
N6frXTpsNVzbQdcS2qlJC9/YgIoJk2KOtWbPJYjNhLixP6Q5D9kCnusSTJV882sF
qV4Wg8y4Z+LoE53MW4LTTLPtW//e5XOsIzstAL81VXQJSdhJWBp/kjbmUZIO8yZ9
HE0XvMnsQybQv0FfQKlERPSZ51eHnlAfV1SoPv10Yy+xUGUJ5lhCLkMaTLTwJUdZ
+gQek9QmRkpQgbLevni3/GcV4clXhB4PY9bpYrrWX1Uu6lzGKAgEJTm4Diup8kyX
HAc/DVL17e8vgg8CAwEAAaOB9DCB8TAfBgNVHSMEGDAWgBStvZh6NLQm9/rEJlTv
A73gJMtUGjAdBgNVHQ4EFgQUU3m/WqorSs9UgOHYm8Cd8rIDZsswDgYDVR0PAQH/
BAQDAgGGMA8GA1UdEwEB/wQFMAMBAf8wEQYDVR0gBAowCDAGBgRVHSAAMEQGA1Ud
HwQ9MDswOaA3oDWGM2h0dHA6Ly9jcmwudXNlcnRydXN0LmNvbS9BZGRUcnVzdEV4
dGVybmFsQ0FSb290LmNybDA1BggrBgEFBQcBAQQpMCcwJQYIKwYBBQUHMAGGGWh0
dHA6Ly9vY3NwLnVzZXJ0cnVzdC5jb20wDQYJKoZIhvcNAQEMBQADggEBAJNl9jeD
lQ9ew4IcH9Z35zyKwKoJ8OkLJvHgwmp1ocd5yblSYMgpEg7wrQPWCcR23+WmgZWn
RtqCV6mVksW2jwMibDN3wXsyF24HzloUQToFJBv2FAY7qCUkDrvMKnXduXBBP3zQ
YzYhBx9G/2CkkeFnvN4ffhkUyWNnkepnB2u0j4vAbkN9w6GAbLIevFOFfdyQoaS8
Le9Gclc1Bb+7RrtubTeZtv8jkpHGbkD4jylW6l/VXxRTrPBPYer3IsynVgviuDQf
Jtl7GQVoP7o81DgGotPmjw7jtHFtQELFhLRAlSv0ZaBIefYdgWOWnU914Ph85I6p
0fKtirOMxyHNwu8=
-----END CERTIFICATE-----'''


URLS = {
    1: {
        'std': {
            'default': {
                'der': 'http://crt.gandi.net/GandiStandardSSLCA.crt',
                'pem':
                    'http://www.gandi.net/static/CAs/GandiStandardSSLCA.pem',
            },
        },
        'pro': {
            'sgc': {
                'pem':
                    'http://www.gandi.net/static/CAs/GandiSGCSSLCA.pem',
            },
            'default': {
                'der': 'http://crt.gandi.net/GandiProSSLCA.crt',
                'pem':
                    'http://www.gandi.net/static/CAs/GandiProSSLCA.pem',
            },
        },
    },
    2: {
        'std': {
            'default': {
                'der': ['http://crt.gandi.net/GandiStandardSSLCA2.crt',
                        'http://crt.usertrust.com/USERTrustRSAAddTrustCA.crt'],
                'pem': [
                    'http://www.gandi.net/static/CAs/GandiStandardSSLCA2.pem'],

            },
        },
        'pro': {
            'default': {
                'der': ['http://crt.gandi.net/GandiProSSLCA2.crt',
                        'http://crt.usertrust.com/USERTrustRSAAddTrustCA.crt'],
                'pem': ['http://www.gandi.net/static/CAs/GandiProSSLCA2.pem',
                        CROSSED_PEM],
            },
        },
    },
}


class Certificate(GandiModule):
    urls = URLS

    """ Module to handle CLI commands.

    $ gandi certificate change-dcv
    $ gandi certificate create
    $ gandi certificate delete
    $ gandi certificate export
    $ gandi certificate info
    $ gandi certificate list
    $ gandi certificate packages
    $ gandi certificate resend-dcv
    $ gandi certificate update
    """

    @classmethod
    def get_latest_valid(cls, hosts):
        """ Retrieve valid certificates by fqdn. """
        certs = cls.list({'status': 'valid', 'items_per_page': 500})
        possible = None

        if not isinstance(hosts, (tuple, list)):
            hosts = [hosts]

        for cert in certs:
            cert_hosts = set([cert['cn']] + cert['altnames'])
            if len(set(hosts) - cert_hosts) == 0:
                if (possible and possible['date_end'] < cert['date_end']
                        or not possible):
                    possible = cert

        return possible

    @classmethod
    def from_cn(cls, common_name):
        """ Retrieve a certificate by its common name. """
        # search with cn
        result_cn = [(cert['id'], [cert['cn']] + cert['altnames'])
                     for cert in cls.list({'status': ['pending', 'valid'],
                                           'items_per_page': 500,
                                           'cn': common_name})]
        # search with altname
        result_alt = [(cert['id'], [cert['cn']] + cert['altnames'])
                      for cert in cls.list({'status': ['pending', 'valid'],
                                            'items_per_page': 500,
                                            'altname': common_name})]

        result = result_cn + result_alt
        ret = {}
        for id_, fqdns in result:
            for fqdn in fqdns:
                ret.setdefault(fqdn, []).append(id_)

        cert_id = ret.get(common_name)
        if not cert_id:
            return

        return cert_id

    @classmethod
    def usable_ids(cls, id, accept_multi=True):
        """ Retrieve id from input which can be an id or a cn."""
        try:
            qry_id = [int(id)]
        except ValueError:
            try:
                qry_id = cls.from_cn(id)
            except Exception:
                qry_id = None

        if not qry_id or not accept_multi and len(qry_id) != 1:
            msg = 'unknown identifier %s' % id
            cls.error(msg)

        return qry_id if accept_multi else qry_id[0]

    @classmethod
    def usable_id(cls, id):
        """ Retrieve id from single input."""
        return cls.usable_ids(id, False)

    @classmethod
    def package_list(cls, options=None):
        """ List possible certificate packages."""
        options = options or {}
        try:
            return cls.safe_call('cert.package.list', options)
        except UsageError as err:
            if err.code == 150020:
                return []
            raise

    __packages__ = None

    @classmethod
    def package_get(cls, package_name):
        if not cls.__packages__:
            cls.__packages__ = dict([(pkg['name'], pkg)
                                     for pkg in cls.package_list()])

        return cls.__packages__.get(package_name)

    @classmethod
    def list(cls, options=None):
        """ List certificates."""
        options = options or {}
        return cls.call('cert.list', options)

    @classmethod
    def info(cls, id):
        """ Display information about a certificate."""
        return cls.call('cert.info', id)

    @classmethod
    def get_package(cls, common_name, type='std', max_altname=None,
                    altnames=None, warranty=None):
        type = type or 'std'

        if max_altname:
            if max_altname < len(altnames):
                cls.echo('You choose --max-altname %s but you have more '
                         'altnames (%s)' % (max_altname, len(altnames)))
                return
        else:
            if '*' in common_name:
                max_altname = 'w'
            elif not altnames:
                max_altname = 1
            else:
                for max_ in [1, 3, 5, 10, 20]:
                    if len(altnames) < max_:
                        max_altname = max_
                        break

                if not max_altname:
                    cls.echo('Too many altnames, max is 20.')
                    return

        pack_filter = 'cert_%s_%s_' % (type, max_altname)
        if warranty:
            pack_filter += '%s_' % (warranty)

        packages = [item['name']
                    for item in cls.package_list()
                    if item['name'].startswith(pack_filter)]

        return packages[0] if packages else None

    @classmethod
    def advice_dcv_method(cls, csr, package, altnames, dcv_method,
                          cert_id=None):
        """ Display dcv_method information. """
        params = {'csr': csr, 'package': package, 'dcv_method': dcv_method}
        if cert_id:
            params['cert_id'] = cert_id
        result = cls.call('cert.get_dcv_params', params)
        if dcv_method == 'dns':
            cls.echo('You have to add these records in your domain zone :')
        cls.echo('\n'.join(result['message']))

    @classmethod
    def change_dcv(cls, oper_id, dcv_method):
        """ Change dcv method."""
        cls.call('cert.change_dcv', oper_id, dcv_method)

    @classmethod
    def resend_dcv(cls, oper_id):
        """ Resend dcv. """
        cls.call('cert.resend_dcv', oper_id)

    @classmethod
    def create(cls, csr, duration, package, altnames=None, dcv_method=None):
        """ Create a new certificate. """
        params = {'csr': csr, 'package': package, 'duration': duration}
        if altnames:
            params['altnames'] = altnames
        if dcv_method:
            params['dcv_method'] = dcv_method

        try:
            result = cls.call('cert.create', params)
        except UsageError:
            params['--dry-run'] = True
            msg = '\n'.join(['%s (%s)' % (err['reason'], err['attr'])
                             for err in cls.call('cert.create', params)])
            cls.error(msg)
            raise

        if dcv_method in ('dns', 'file'):
            cls.advice_dcv_method(csr, package, altnames, dcv_method,
                                  cert_id=result['params'].get('cert_id'))
        return result

    @classmethod
    def update(cls, cert_id, csr, private_key, country, state, city,
               organisation, branch, altnames, dcv_method):
        """ Update a certificate. """
        cert = cls.info(cert_id)
        if cert['status'] != 'valid':
            cls.error('The certificate must be in valid status to be updated.')
            return

        common_name = cert['cn']

        csr = cls.process_csr(common_name, csr, private_key, country, state,
                              city, organisation, branch)

        if not csr:
            return

        params = {'csr': csr}
        if altnames:
            params['altnames'] = []
            for altname in altnames:
                params['altnames'].extend(altname.split(','))
        if dcv_method:
            params['dcv_method'] = dcv_method

        try:
            result = cls.call('cert.update', cert_id, params)
        except UsageError:
            params['--dry-run'] = True
            msg = cls.call('cert.update', cert_id, params)
            if msg:
                cls.error(str(msg))
            raise

        return result

    @staticmethod
    def private_key(common_name):
        return common_name.replace('*.', 'wildcard.') + '.key'

    @classmethod
    def gen_pk(cls, common_name, private_key):
        if private_key:
            cmd = 'openssl req -new -key %(key)s -out %(csr)s -subj "%(subj)s"'
            if not os.path.exists(private_key):
                content = private_key
                private_key = cls.private_key(common_name)
                with open(private_key, 'w') as fhandle:
                    fhandle.write(content)
        else:
            private_key = cls.private_key(common_name)
            # TODO check if it exists
            cmd = ('openssl req -new -newkey rsa:2048 -sha256 -nodes '
                   '-out %(csr)s -keyout %(key)s -subj "%(subj)s"')
        return cmd, private_key

    @classmethod
    def create_csr(cls, common_name, private_key=None, params=None):
        """ Create CSR. """
        params = params or []

        params = [(key, val) for key, val in params if val]
        subj = '/' + '/'.join(['='.join(value) for value in params])

        cmd, private_key = cls.gen_pk(common_name, private_key)

        if private_key.endswith('.crt') or private_key.endswith('.key'):
            csr_file = re.sub(r'\.(crt|key)$', '.csr', private_key)
        else:
            csr_file = private_key + '.csr'

        cmd = cmd % {'csr': csr_file, 'key': private_key, 'subj': subj}
        result = cls.execute(cmd)
        if not result:
            cls.echo('CSR creation failed')
            cls.echo(cmd)
            return

        return csr_file

    @classmethod
    def get_common_name(cls, csr):
        """ Read information from CSR. """
        from tempfile import NamedTemporaryFile
        fhandle = NamedTemporaryFile()
        fhandle.write(csr.encode('latin1'))
        fhandle.flush()
        output = cls.exec_output('openssl req -noout -subject -in %s' %
                                 fhandle.name)
        if not output:
            return

        common_name = output.split('=')[-1].strip()
        fhandle.close()
        return common_name

    @classmethod
    def process_csr(cls, common_name, csr=None, private_key=None,
                    country=None, state=None, city=None, organisation=None,
                    branch=None):
        """ Create a PK and a CSR if needed."""
        if csr:
            if branch or organisation or city or state or country:
                cls.echo('Following options are only used to generate'
                         ' the CSR.')
        else:
            params = (('CN', common_name),
                      ('OU', branch),
                      ('O', organisation),
                      ('L', city),
                      ('ST', state),
                      ('C', country))
            params = [(key, val) for key, val in params if val]
            csr = cls.create_csr(common_name, private_key, params)

        if csr and os.path.exists(csr):
            with open(csr) as fcsr:
                csr = fcsr.read()

        return csr

    @classmethod
    def pretty_format_cert(cls, cert):
        """ Pretty display of a certificate."""
        crt = cert.get('cert')
        if crt:
            crt = ('-----BEGIN CERTIFICATE-----\n' +
                   '\n'.join([crt[index * 64:(index + 1) * 64]
                              for index in range(int(len(crt) / 64) + 1)]).rstrip('\n') +  # noqa
                   '\n-----END CERTIFICATE-----')
        return crt

    @classmethod
    def delete(cls, cert_id, background=False):
        """ Delete a certificate."""
        result = cls.call('cert.delete', cert_id)

        if background:
            return result

        cls.echo("Deleting your certificate.")
        cls.display_progress(result)
        cls.echo('Your certificate %s has been deleted.' % cert_id)

        return result



from .compat import unittest
from .compat import mock


class TestCase(unittest.TestCase):

    def test_main(self):
        cli = mock.Mock()
        with mock.patch('gandi.cli.core.cli.cli', cli):
            from gandi.cli.__main__ import main
            main()
            cli.assert_called_once_with(obj={})



""" Contact namespace commands. """

import click
import time
import webbrowser

# define unicode for python3
try:
    unicode
except NameError:
    unicode = str

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import randomstring
from gandi.cli.core.params import pass_gandi


FIELDS = (('type', 'Choose your contact type',
           {'valid': ((0, 'individual'),
                      (1, 'company'),
                      (2, 'association'),
                      (3, 'public body'),
                      (4, 'reseller')),
            'convert': int}),
          ('given', 'What is your first name', None),
          ('family', 'What is your last name', None),
          ('orgname', 'What is your company name',
           {'display': lambda contact_: contact_['type'] != 0}),
          ('email', 'What is your email address', None),
          ('streetaddr', 'What is your street address', None),
          ('zip', 'What is your zipcode', None),
          ('city', 'Which city', None),
          ('country', 'Which country', None),
          ('phone', 'What is your telephone number', None))


FIELDS_POSITION = dict([(j, i) for i, j in enumerate(
                        [field[0] for field in FIELDS])])


def ask_field(gandi, contact, field, label, checks):
    valid = display = None
    convert = unicode

    if checks:
        valid = checks.get('valid')
        convert = checks.get('convert', unicode)
        display = checks.get('display')

    if display and not display(contact):
        return

    if not valid:
        contact[field] = convert(click.prompt(label))
    elif isinstance(valid, (tuple, list)):
        valid_keys = [unicode(val) for val, _ in valid]
        value = None
        gandi.echo(label)
        while value not in valid_keys:
            for key, val in valid:
                gandi.echo('%s- %s' % (key, val))
            value = click.prompt('')
        contact[field] = convert(value)


@cli.group(name='contact')
@pass_gandi
def contact(gandi):
    """Commands related to contacts."""


@contact.command()
@pass_gandi
def create(gandi):
    """ Create a new contact.
    """
    contact = {}

    for field, label, checks in FIELDS:
        ask_field(gandi, contact, field, label, checks)

    default_pwd = randomstring(16)
    contact['password'] = click.prompt('Please enter your password',
                                       hide_input=True,
                                       confirmation_prompt=True,
                                       default=default_pwd)

    result = True
    while result:
        result = gandi.contact.create_dry_run(contact)

        # display errors
        for err in result:
            gandi.echo(err['reason'])
            field = err['field']

            if field not in FIELDS_POSITION:
                return

            desc = FIELDS[FIELDS_POSITION.get(field)]
            ask_field(gandi, contact, *desc)

    result = gandi.contact.create(contact)
    handle = result['handle']

    gandi.echo('Please activate you public api access from gandi website, and '
               'get the apikey.')
    gandi.echo('Your handle is %s, and the password is the one you defined.' %
               handle)

    # open new browser window
    webbrowser.open('https://www.gandi.net/admin/api_key')
    # just to avoid missing the next question in webbrowser stderr
    time.sleep(1)

    # get the apikey from shell
    apikey = None
    while not apikey:
        apikey = click.prompt('What is your production apikey')

    caller = gandi.get('api.key')
    # save apikey in the conf if none defined else display an help on how to
    # use it
    if caller:
        gandi.echo('You already have an apikey defined, if you want to use the'
                   ' newly created contact, use the env var : ')
        gandi.echo('export API_KEY=%s' % apikey)
    else:
        gandi.echo('Will save your apikey into the config file.')
        gandi.configure(True, 'api.key', apikey)

    return handle



# -*- coding: utf-8 -*-
""" GandiCLI class declaration and initialization. """

import os
import os.path
import inspect
import platform
from functools import update_wrapper

import click

from .base import GandiContextHelper
from gandi.cli import __version__

try:
    use_man_epilog = platform.system() == 'Linux'
except Exception:
    pass


# XXX: dirty hack of click help command to allow short help -h
def add_help_option(self):
    """Add a help option to the command."""
    click.help_option(*('--help', '-h'))(self)


click.Command.add_help_option = add_help_option

# XXX: patch each command with an epilog
if use_man_epilog:
    def format_epilog(self, ctx, formatter):
        """Writes the epilog into the formatter if it exists."""
        self.epilog = 'For detailed documentation, use `man gandi`.'
        formatter.write_paragraph()
        with formatter.indentation():
            formatter.write_text(self.epilog)
    click.Command.format_epilog = format_epilog


def compatcallback(f):
    """ Compatibility callback decorator for older click version.

    Click 1.0 does not have a version string stored, so we need to
    use getattr here to be safe.
    """
    if getattr(click, '__version__', '0.0') >= '2.0':
        return f
    return update_wrapper(lambda ctx, value: f(ctx, None, value), f)


class GandiCLI(click.Group):

    """ Gandi command line utility.

    All CLI commands have a documented help

    $ gandi <command> --help

    """

    def __init__(self, help=None):
        """ Initialize CLI command line."""
        @compatcallback
        def set_debug(ctx, param, value):
            ctx.obj['verbose'] = value

        @compatcallback
        def get_version(ctx, param, value):
            if value:
                print(('Gandi CLI %s\n\n'
                       'Copyright: © 2014-2018 Gandi S.A.S.\n'
                       'License: GPL-3' % __version__))
                ctx.exit()

        if help is None:
            help = inspect.getdoc(self)

        click.Group.__init__(self, help=help, params=[
            click.Option(['-v'],
                         help='Enable or disable verbose mode. Use multiple '
                              'time for higher level of verbosity: -v, -vv',
                         count=True, metavar='',
                         default=0, callback=set_debug),
            click.Option(['--version'],
                         help='Display version.',
                         is_flag=True,
                         default=False, callback=get_version)

        ])

    def format_commands(self, ctx, formatter):
        """Extra format methods for multi methods that adds all the commands
        after the options.

        Display custom help for all subcommands.
        """
        rows = []

        all_cmds = self.list_all_commands(ctx)
        for cmd_name in sorted(all_cmds):
            cmd = all_cmds[cmd_name]
            help = cmd.short_help or ''
            rows.append((cmd_name, help))

        if rows:
            with formatter.section('Commands'):
                formatter.write_dl(rows)

    def list_sub_commmands(self, cmd_name, cmd):
        """Return all commands for a group"""
        ret = {}
        if isinstance(cmd, click.core.Group):
            for sub_cmd_name in cmd.commands:
                sub_cmd = cmd.commands[sub_cmd_name]
                sub = self.list_sub_commmands(sub_cmd_name, sub_cmd)
                if sub:
                    if isinstance(sub, dict):
                        for n, c in sub.items():
                            ret['%s %s' % (cmd_name, n)] = c
                    else:
                        ret['%s %s' % (cmd_name, sub[0])] = sub[1]
        elif isinstance(cmd, click.core.Command):
            return (cmd.name, cmd)
        return ret

    def list_all_commands(self, ctx):
        ret = {}
        for cmd_name in self.commands:
            cmd = self.commands[cmd_name]
            sub = self.list_sub_commmands(cmd_name, cmd)
            if sub:
                if isinstance(sub, tuple):
                    ret[sub[0]] = sub[1]
                else:
                    ret.update(sub)
        return ret

    def load_commands(self):
        """ Load cli commands from submodules. """
        command_folder = os.path.join(os.path.dirname(__file__),
                                      '..', 'commands')
        command_dirs = {
            'gandi.cli': command_folder
        }

        if 'GANDICLI_PATH' in os.environ:
            for _path in os.environ.get('GANDICLI_PATH').split(':'):
                # remove trailing separator if any
                path = _path.rstrip(os.sep)
                command_dirs[os.path.basename(path)] = os.path.join(path,
                                                                    'commands')

        for module_basename, dir in list(command_dirs.items()):
            for filename in sorted(os.listdir(dir)):
                if filename.endswith('.py') and '__init__' not in filename:
                    submod = filename[:-3]
                    module_name = module_basename + '.commands.' + submod
                    __import__(module_name, fromlist=[module_name])

    def invoke(self, ctx):
        """ Invoke command in context. """
        ctx.obj = GandiContextHelper(verbose=ctx.obj['verbose'])
        click.Group.invoke(self, ctx)


cli = GandiCLI()
cli.load_commands()



""" Contains output methods used by commands.

Also custom exceptions and method to generate a random string.
"""
import sys
import time
from datetime import datetime

import json
from click.formatting import measure_table
from click import ClickException

from .ascii_sparks import sparks


class MissingConfiguration(Exception):

    """ Raise when configuration if missing. """

    def __init__(self, errors):
        """ Initialize exception."""
        self.errors = errors


class DuplicateResults(Exception):

    """ Raise when multiple results are found."""

    def __init__(self, errors):
        """ Initialize exception."""
        self.errors = errors


class DomainNotAvailable(Exception):

    """ Raise when domain is not available. """

    def __init__(self, errors):
        """ Initialize exception."""
        self.errors = errors


class DatacenterClosed(ClickException):

    """Raise when datacenter is closed: ALL"""

    def __init__(self, message):
        """ Initialize exception."""
        self.message = message


class DatacenterLimited(Exception):

    """Raise when datacenter will soon be closed: NEW"""

    def __init__(self, date):
        """ Initialize exception."""
        self.date = date


class MigrationNotFinalized(ClickException):

    """Raise when VM migration does not need to be finalized."""

    def __init__(self, message):
        """ Initialize exception."""
        self.message = message


def format_list(data):
    """ Remove useless characters to output a clean list."""
    if isinstance(data, (list, tuple)):
        to_clean = ['[', ']', '(', ')', "'"]
        for item in to_clean:
            data = str(data).replace("u\"", "\"").replace("u\'", "\'")
            data = str(data).replace(item, '')
    return data


def display_rows(gandi, rows, has_header=True):
    col_len = measure_table(rows)
    formatting = ' | '.join(['%-' + str(l) + 's' for l in col_len])

    if has_header:
        header = rows.pop(0)
        gandi.echo(formatting % tuple(header))
        gandi.echo('-+-'.join(['-' * l for l in col_len]))

    for row in rows:
        gandi.echo(formatting % tuple(row))


def output_line(gandi, key, val, justify):
    """ Base helper to output a key value using left justify."""
    msg = ('%%-%ds:%%s' % justify) % (key, (' %s' % val) if val else '')
    gandi.echo(msg)


def output_generic(gandi, data, output_keys, justify=10):
    """ Generic helper to output info from a data dict."""
    for key in output_keys:
        if key in data:
            output_line(gandi, key, data[key], justify)


def output_account(gandi, account, output_keys, justify=17):
    """ Helper to output an account information."""
    output_generic(gandi, account, output_keys, justify)

    if 'prepaid' in output_keys:
        prepaid = '%s %s' % (account['prepaid_info']['amount'],
                             account['prepaid_info']['currency'])
        output_line(gandi, 'prepaid', prepaid, justify)

    if 'credit' in output_keys:
        output_line(gandi, 'credits', None, justify)
        available = account.get('credits')
        output_line(gandi, '        available', available, justify)
        # sometimes rating is returning nothing
        usage_str = left_str = 'not available'
        usage = account.get('credit_usage', 0)
        left = account.get('left')
        if usage:
            usage_str = '%d/h' % usage

            years, months, days, hours = left
            left_str = ('%d year(s) %d month(s) %d day(s) %d hour(s)' %
                        (years, months, days, hours))

        output_line(gandi, '        usage', usage_str, justify)
        output_line(gandi, '        time left', left_str, justify)


def output_vm(gandi, vm, datacenters, output_keys, justify=10):
    """ Helper to output a vm information."""
    output_generic(gandi, vm, output_keys, justify)

    if 'datacenter' in output_keys:
        for dc in datacenters:
            if dc['id'] == vm['datacenter_id']:
                dc_name = dc.get('dc_code', dc.get('iso', ''))
                break

        output_line(gandi, 'datacenter', dc_name, justify)

    if 'ip' in output_keys:
        for iface in vm['ifaces']:
            gandi.separator_line()
            output_line(gandi, 'bandwidth', iface['bandwidth'], justify)

            for ip in iface['ips']:
                ip_addr = ip['ip']

                output_line(gandi, 'ip%s' % ip['version'], ip_addr, justify)


def output_metric(gandi, metrics, key, justify=10):
    """ Helper to output metrics."""
    for metric in metrics:
        key_name = metric[key].pop()
        values = [point.get('value', 0) for point in metric['points']]
        graph = sparks(values) if max(values) else ''
        # need to encode in utf-8 to work in python2.X
        if sys.version_info < (2, 8):
            graph = graph.encode('utf-8')
        output_line(gandi, key_name, graph, justify)


def output_vhost(gandi, vhost, paas, output_keys, justify=14):
    """ Helper to output a vhost information."""
    output_generic(gandi, vhost, output_keys, justify)

    if 'paas_name' in output_keys:
        output_line(gandi, 'paas_name', paas, justify)


def output_paas(gandi, paas, datacenters, vhosts, output_keys, justify=11):
    """ Helper to output a paas information."""
    output_generic(gandi, paas, output_keys, justify)

    if 'sftp_server' in output_keys:
        output_line(gandi, 'sftp_server', paas['ftp_server'], justify)

    if 'vhost' in output_keys:
        for entry in vhosts:
            output_line(gandi, 'vhost', entry, justify)

    if 'dc' in output_keys:
        dc_name = paas['datacenter'].get('dc_code',
                                         paas['datacenter'].get('iso', ''))
        output_line(gandi, 'datacenter', dc_name, justify)

    if 'df' in paas:
        df = paas['df']
        total = df['free'] + df['used']
        if total:
            disk_used = '%.1f%%' % (df['used'] * 100 / total)
            output_line(gandi, 'quota used', disk_used, justify)

    if 'snapshot' in output_keys:
        val = None
        if paas['snapshot_profile']:
            val = paas['snapshot_profile']['name']
        output_line(gandi, 'snapshot', val, justify)

    if 'cache' in paas:
        cache = paas['cache']
        total = cache['hit'] + cache['miss'] + cache['not'] + cache['pass']
        if total:
            output_line(gandi, 'cache', None, justify)
            for key in sorted(cache):
                str_value = '%.1f%%' % (cache[key] * 100 / total)
                output_sub_line(gandi, key, str_value, 5)


def output_image(gandi, image, datacenters, output_keys, justify=14,
                 warn_deprecated=True):
    """ Helper to output a disk image."""

    for key in output_keys:
        if key in image:
            if (key == 'label' and image['visibility'] == 'deprecated' and
                    warn_deprecated):
                image[key] = '%s /!\ DEPRECATED' % image[key]
            output_line(gandi, key, image[key], justify)

    dc_name = 'Nowhere'

    if 'dc' in output_keys:
        for dc in datacenters:
            if dc['id'] == image['datacenter_id']:
                dc_name = dc.get('dc_code', dc.get('iso', ''))

                break

        output_line(gandi, 'datacenter', dc_name, justify)


def output_kernels(gandi, flavor, name_list, justify=14):
    """ Helper to output kernel flavor versions."""
    output_line(gandi, 'flavor', flavor, justify)
    for name in name_list:
        output_line(gandi, 'version', name, justify)


def output_datacenter(gandi, datacenter, output_keys, justify=14):
    """ Helper to output datacenter information."""
    output_generic(gandi, datacenter, output_keys, justify)

    if 'dc_name' in output_keys:
        output_line(gandi, 'datacenter', datacenter['name'], justify)

    if 'status' in output_keys:
        deactivate_at = datacenter.get('deactivate_at')
        if deactivate_at:
            output_line(gandi, 'closing on',
                        deactivate_at.strftime('%d/%m/%Y'), justify)

        closing = []
        iaas_closed_for = datacenter.get('iaas_closed_for')
        if iaas_closed_for == 'ALL':
            closing.append('vm')

        paas_closed_for = datacenter.get('paas_closed_for')
        if paas_closed_for == 'ALL':
            closing.append('paas')

        if closing:
            output_line(gandi, 'closed for', ', '.join(closing), justify)


def output_cmdline(gandi, cmdline, justify=14):
    args = []
    for key in sorted(cmdline, reverse=True):
        if isinstance(cmdline[key], bool):
            args.append(key)
        else:
            args.append('%s=%s' % (key, cmdline[key]))
    output_line(gandi, 'cmdline', ' '.join(args), justify)


def output_disk(gandi, disk, datacenters, vms, profiles, output_keys,
                justify=10):
    """ Helper to output a disk."""
    output_generic(gandi, disk, output_keys, justify)

    if 'kernel' in output_keys and disk.get('kernel_version'):
        output_line(gandi, 'kernel', disk['kernel_version'], justify)

    if 'cmdline' in output_keys and disk.get('kernel_cmdline'):
        output_cmdline(gandi, disk.get('kernel_cmdline'), justify)

    if 'dc' in output_keys:
        dc_name = None
        for dc in datacenters:
            if dc['id'] == disk['datacenter_id']:
                dc_name = dc.get('dc_code', dc.get('iso', ''))
                break

        if dc_name:
            output_line(gandi, 'datacenter', dc_name, justify)

    if 'vm' in output_keys:
        for vm_id in disk['vms_id']:
            vm_name = vms.get(vm_id, {}).get('hostname')
            if vm_name:
                output_line(gandi, 'vm', vm_name, justify)

    if 'profile' in output_keys and disk.get('snapshot_profile'):
        output_line(gandi, 'profile', disk['snapshot_profile']['name'],
                    justify)
    elif 'profile' in output_keys and disk.get('snapshot_profile_id'):
        for profile in profiles:
            if profile['id'] == disk['snapshot_profile_id']:
                output_line(gandi, 'profile', profile['name'], justify)
                break


def output_sshkey(gandi, sshkey, output_keys, justify=12):
    """ Helper to output an ssh key information."""
    output_generic(gandi, sshkey, output_keys, justify)


def output_snapshot_profile(gandi, profile, output_keys, justify=13):
    """ Helper to output a snapshot_profile."""
    schedules = 'schedules' in output_keys
    if schedules:
        output_keys.remove('schedules')
    output_generic(gandi, profile, output_keys, justify)

    if schedules:
        schedule_keys = ['name', 'kept_version']
        for schedule in profile['schedules']:
            gandi.separator_line()
            output_generic(gandi, schedule, schedule_keys, justify)


def output_contact_info(gandi, data, output_keys, justify=10):
    """Helper to output chosen contacts info."""
    for key in output_keys:
        if data[key]:
            output_line(gandi, key, data[key]['handle'], justify)


def output_cert_oper(gandi, oper, justify=12):
    output_generic(gandi, oper, ['type', 'step'], justify)
    params = dict(oper['params'])
    params['fqdns'] = ', '.join(params.get('fqdns', []))
    output = ['inner_step', 'package_name', 'dcv_method']
    if params['fqdns']:
        output.append('fqdns')
    output_generic(gandi, params, output, justify)


def output_cert(gandi, cert, output_keys, justify=13):
    """Helper to output a certificate information."""
    output = list(output_keys)

    display_altnames = False
    if 'altnames' in output:
        display_altnames = True
        output.remove('altnames')

    display_output = False
    if 'cert' in output:
        display_output = True
        output.remove('cert')

    output_generic(gandi, cert, output, justify)

    if display_output:
        crt = gandi.certificate.pretty_format_cert(cert)
        if crt:
            output_line(gandi, 'cert', '\n' + crt, justify)

    if display_altnames:
        for altname in cert['altnames']:
            output_line(gandi, 'altname', altname, justify)


def output_vlan(gandi, vlan, datacenters, output_keys, justify=10):
    """ Helper to output a vlan information."""
    output_generic(gandi, vlan, output_keys, justify)

    if 'dc' in output_keys:
        for dc in datacenters:
            if dc['id'] == vlan.get('datacenter_id',
                                    vlan.get('datacenter', {}).get('id')):
                dc_name = dc.get('dc_code', dc.get('iso', ''))
                break

        output_line(gandi, 'datacenter', dc_name, justify)


def output_iface(gandi, iface, datacenters, vms, output_keys, justify=10):
    """ Helper to output an iface information."""
    output_generic(gandi, iface, output_keys, justify)

    if 'vm' in output_keys:
        vm_name = vms.get(iface['vm_id'], {}).get('hostname')
        if vm_name:
            output_line(gandi, 'vm', vm_name, justify)

    if 'dc' in output_keys:
        for dc in datacenters:
            if dc['id'] == iface.get('datacenter_id',
                                     iface.get('datacenter', {}).get('id')):
                dc_name = dc.get('dc_code', dc.get('iso', ''))
                break

        output_line(gandi, 'datacenter', dc_name, justify)

    if 'vlan_' in output_keys:
        vlan = iface.get('vlan') or {}
        output_line(gandi, 'vlan', vlan.get('name', '-'), justify)


def output_ip(gandi, ip, datacenters, vms, ifaces, output_keys, justify=11):
    """ Helper to output an ip information."""
    output_generic(gandi, ip, output_keys, justify)

    if 'type' in output_keys:
        iface = ifaces.get(ip['iface_id'])
        type_ = 'private' if iface.get('vlan') else 'public'
        output_line(gandi, 'type', type_, justify)
        if type_ == 'private':
            output_line(gandi, 'vlan', iface['vlan']['name'], justify)

    if 'vm' in output_keys:
        iface = ifaces.get(ip['iface_id'])
        vm_id = iface.get('vm_id')
        if vm_id:
            vm_name = vms.get(vm_id, {}).get('hostname')
            if vm_name:
                output_line(gandi, 'vm', vm_name, justify)

    if 'dc' in output_keys:
        for dc in datacenters:
            if dc['id'] == ip.get('datacenter_id',
                                  ip.get('datacenter', {}).get('id')):
                dc_name = dc.get('dc_code', dc.get('iso', ''))
                break

        output_line(gandi, 'datacenter', dc_name, justify)


def randomstring(prefix=None):
    """ Helper to generate a random string, used for temporary hostnames."""
    if not prefix:
        prefix = 'tmp'
    return '%s%s' % (prefix, str(int(time.time())))


def output_list(gandi, val):
    """Helper to generate a beautiful list."""
    for element in val:
        gandi.echo(element)


def date_handler(obj):
    """ Serialize date for json output """
    return obj.isoformat() if hasattr(obj, 'isoformat') else str(obj)


def output_json(gandi, format, value):
    """ Helper to show json output """
    if format == 'json':
        gandi.echo(json.dumps(value, default=date_handler, sort_keys=True))
    elif format == 'pretty-json':
        gandi.echo(json.dumps(value, default=date_handler, sort_keys=True,
                              indent=2, separators=(',', ': ')))


def output_sub_line(gandi, key, val, justify):
    """ Base helper to output a key value using left justify."""
    msg = ('\t%%-%ds:%%s' % justify) % (key, (' %s' % val) if val else '')
    gandi.echo(msg)


def output_sub_generic(gandi, data, output_keys, justify=10):
    """ Generic helper to output info from a data dict."""
    for key in output_keys:
        if key in data:
            output_sub_line(gandi, key, data[key], justify)


def output_service(gandi, service, status, justify=10):
    """ Helper to output a status service information."""
    output_line(gandi, service, status, justify)


def output_hostedcert(gandi, hcert, output_keys, justify=12):
    output_keys = list(output_keys)
    fqdns = 'fqdns' in output_keys
    vhosts = 'vhosts' in output_keys

    if fqdns:
        output_keys.pop(output_keys.index('fqdns'))

    if vhosts:
        output_keys.pop(output_keys.index('vhosts'))

    output_generic(gandi, hcert, output_keys, justify)

    if fqdns:
        for fqdn in hcert['fqdns']:
            gandi.separator_sub_line()
            output_sub_line(gandi, 'fqdn', fqdn['name'], 10)

    if vhosts:
        for vhost in hcert['related_vhosts']:
            gandi.separator_sub_line()
            output_sub_line(gandi, 'vhost', vhost['name'], 10)
            output_sub_line(gandi, 'type', vhost['type'], 10)


def output_domain(gandi, domain, output_keys, justify=12):
    """ Helper to output a domain information."""
    if 'nameservers' in domain:
        domain['nameservers'] = format_list(domain['nameservers'])

    if 'services' in domain:
        domain['services'] = format_list(domain['services'])

    if 'tags' in domain:
        domain['tags'] = format_list(domain['tags'])

    output_generic(gandi, domain, output_keys, justify)

    if 'created' in output_keys:
        output_line(gandi, 'created', domain['date_created'], justify)

    if 'expires' in output_keys:
        date_end = domain.get('date_registry_end')
        if date_end:
            days_left = (date_end - datetime.now()).days
        output_line(gandi, 'expires',
                    '%s (in %d days)' % (date_end, days_left),
                    justify)

    if 'updated' in output_keys:
        output_line(gandi, 'updated', domain['date_updated'], justify)


def output_mailbox(gandi, mailbox, output_keys, justify=16):
    """ Helper to output a mailbox information."""
    quota = 'quota' in output_keys
    responder = 'responder' in output_keys

    if quota:
        output_keys.pop(output_keys.index('quota'))

    if responder:
        output_keys.pop(output_keys.index('responder'))

    if 'aliases' in output_keys:
        mailbox['aliases'] = sorted(mailbox['aliases'])

    output_generic(gandi, mailbox, output_keys, justify)

    if 'fallback' in output_keys:
        output_line(gandi, 'fallback email', mailbox['fallback_email'],
                    justify)

    if quota:
        granted = mailbox['quota']['granted']
        if mailbox['quota']['granted'] == 0:
            granted = 'unlimited'
        output_line(gandi, 'quota usage',
                    '%s KiB / %s' % (mailbox['quota']['used'], granted),
                    justify)

    if responder:
        responder_status = 'yes' if mailbox['responder']['active'] else 'no'
        output_line(gandi, 'responder active', responder_status, justify)
        output_line(gandi, 'responder text', mailbox['responder']['text'],
                    justify)


def output_forward(gandi, domain, forward, justify=14):
    """ Helper to output a mail forward information."""
    for dest in forward['destinations']:
        output_line(gandi, forward['source'], dest, justify)


def output_dns_records(gandi, records, output_keys, justify=12):
    """ Helper to output a dns records information."""
    for key in output_keys:
        real_key = 'rrset_%s' % key
        if real_key in records:
            val = records[real_key]
            if key == 'values':
                val = format_list(records[real_key])
            output_line(gandi, key, val, justify)



""" Hosted certificate namespace commands. """

import os
import click

# define basestring for python3
try:
    basestring
except NameError:
    basestring = (str, bytes)

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import output_hostedcert
from gandi.cli.core.params import pass_gandi

type_list = list


@cli.group(name='certstore')
@pass_gandi
def certstore(gandi):
    """Commands related to certificate store."""


@certstore.command()
@click.option('--id', help='Display ids.', is_flag=True)
@click.option('--vhosts', help='Display related vhosts.', is_flag=True)
@click.option('--dates', help='Display dates.', is_flag=True)
@click.option('--fqdns', help='Display fqdns.', is_flag=True)
@click.option('--limit', help='Limit number of results.', default=100,
              show_default=True)
@pass_gandi
def list(gandi, id, vhosts, dates, fqdns, limit):
    """ List hosted certificates. """
    justify = 10
    options = {'items_per_page': limit, 'state': 'created'}

    output_keys = []

    if id:
        output_keys.append('id')

    output_keys.append('subject')

    if dates:
        output_keys.extend(['date_created', 'date_expire'])
        justify = 12

    if fqdns:
        output_keys.append('fqdns')

    if vhosts:
        output_keys.append('vhosts')

    result = gandi.hostedcert.list(options)

    for num, hcert in enumerate(result):
        if num:
            gandi.separator_line()

        if fqdns or vhosts:
            hcert = gandi.hostedcert.info(hcert['id'])

        output_hostedcert(gandi, hcert, output_keys, justify)

    return result


@certstore.command()
@click.argument('resource', nargs=-1, required=True)
@pass_gandi
def info(gandi, resource):
    """ Display information about a hosted certificate.

    Resource can be a FQDN or an ID
    """
    output_keys = ['id', 'subject', 'date_created', 'date_expire',
                   'fqdns', 'vhosts']

    result = gandi.hostedcert.infos(resource)
    for num, hcert in enumerate(result):
        if num:
            gandi.separator_line()
        output_hostedcert(gandi, hcert, output_keys)

    return result


@certstore.command()
@click.option('--private-key', '--pk', required=True,
              help='Private key used to generate this CRT.')
@click.option('--certificate', '--crt', required=False,
              help='The certificate.')
@click.option('--certificate-id', '--crt-id', type=click.INT, required=False,
              help='The certificate.')
@pass_gandi
def create(gandi, private_key, certificate, certificate_id):
    """ Create a new hosted certificate. """
    if not certificate and not certificate_id:
        gandi.echo('One of --certificate or --certificate-id is needed.')
        return
    if certificate and certificate_id:
        gandi.echo('Only one of --certificate or --certificate-id is needed.')

    if os.path.isfile(private_key):
        with open(private_key) as fhandle:
            private_key = fhandle.read()

    if certificate:
        if os.path.isfile(certificate):
            with open(certificate) as fhandle:
                certificate = fhandle.read()
    else:
        cert = gandi.certificate.info(certificate_id)
        certificate = gandi.certificate.pretty_format_cert(cert)

    result = gandi.hostedcert.create(private_key, certificate)

    output_keys = ['id', 'subject', 'date_created', 'date_expire',
                   'fqdns', 'vhosts']

    output_hostedcert(gandi, result, output_keys)

    return result


@certstore.command()
@click.argument('resource', nargs=-1, required=True)
@click.option('--force', '-f', is_flag=True,
              help='This is a dangerous option that will cause CLI to continue'
                   ' without prompting. (default=False).')
@pass_gandi
def delete(gandi, resource, force):
    """ Delete a hosted certificate.

    Resource can be a FQDN or an ID
    """
    infos = gandi.hostedcert.infos(resource)
    if not infos:
        return

    if not force:
        proceed = click.confirm('Are you sure to delete the following hosted '
                                'certificates ?\n' +
                                '\n'.join(['%s: %s' % (res['id'],
                                                       res['subject'])
                                           for res in infos]) + '\n')
        if not proceed:
            return

    for res in infos:
        gandi.hostedcert.delete(res['id'])



#!/usr/bin/python

import os
import socket
import select
import subprocess
import sys
import time


class FdPipe:

    """Connect two pairs of file objects"""

    def __init__(self, in0, out0, in1, out1):
        self.poller = select.poll()
        self.fd_map = {
            out0.fileno(): in1.fileno(),
            out1.fileno(): in0.fileno()
        }

        for fd in self.fd_map:
            self.select_for_read(fd)

        self.out_buf = {}

    def select_for_flush(self, fd):
        self.poller.register(fd, select.POLLERR | select.POLLHUP |
                             select.POLLIN | select.POLLOUT)

    def select_for_write(self, fd):
        self.poller.register(fd, select.POLLERR | select.POLLHUP |
                             select.POLLIN | select.POLLOUT)

    def select_for_read(self, fd):
        self.poller.register(fd, select.POLLERR | select.POLLHUP |
                             select.POLLIN)

    def do_write(self, outfd):
        if not self.out_buf[outfd]:
            return

        (rl, wl, el) = select.select([], [outfd], [outfd])
        if outfd in wl:
            length = os.write(outfd, self.out_buf[outfd])
            self.out_buf[outfd] = self.out_buf[outfd][length:]
            if not self.out_buf[outfd]:
                self.select_for_read(outfd)
                del self.out_buf[outfd]
        elif outfd in el:
            raise Exception('could not flush fd')

    def queue_write(self, outfd, data):
        self.out_buf[outfd] = self.out_buf.setdefault(outfd, '') + data
        self.select_for_write(outfd)

    def flush_outputs(self):
        while self.out_buf:
            for outfd in self.out_buf:
                try:
                    self.do_write(outfd)
                except OSError:
                    return

    def one_loop(self):
        ret = True
        for (fd, ev) in self.poller.poll(1000):
            if ev & (select.POLLERR | select.POLLHUP):
                self.flush_outputs()
                self.poller.unregister(fd)
                ret = False
            if ev & select.POLLIN:
                data = os.read(fd, 4096)
                if not data:
                    os.close(self.fd_map[fd])
                    ret = False
                self.queue_write(self.fd_map[fd], data)
            if ev & select.POLLOUT:
                self.do_write(fd)
        return ret


def scp(addr, user, local_path, remote_path, local_key=None):
    scp_call = ['scp', local_path,
                '%s@[%s]:%s' % (user, addr, remote_path)]

    if local_key:
        scp_call.insert(1, local_key)
        scp_call.insert(1, '-i')

    subprocess.call(scp_call)


def tcp4_to_unix(local_port, unix_path):
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM,
                           socket.IPPROTO_TCP)
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    try:
        server.bind(('127.0.0.1', local_port))
    except socket.error as e:
        sys.stderr.write('remote cant grab port %d\n' % local_port)
        # let other end time to connect to maintain ssh up
        time.sleep(10)
        sys.exit(0)
    server.listen(32)

    while True:
        (rl, wl, el) = select.select([server], [], [server], 1)
        if server in rl:
            (client, _) = server.accept()
            if not os.fork():
                unix = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                try:
                    unix.connect(unix_path)
                except socket.error as e:
                    print('Unable to grab %s: %s' % (unix_path, e))
                pipe = FdPipe(client, client, unix, unix)
                while pipe.one_loop():
                    pass
                return
            client.close()
        try:
            os.waitpid(-1, os.WNOHANG)
        except OSError:
            pass


def find_port(addr, user):
    """Find local port in existing tunnels"""
    import pwd
    home = pwd.getpwuid(os.getuid()).pw_dir
    for name in os.listdir('%s/.ssh/' % home):
        if name.startswith('unixpipe_%s@%s_' % (user, addr,)):
            return int(name.split('_')[2])


def new_port():
    """Find a free local port and allocate it"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
    for i in range(12042, 16042):
        try:
            s.bind(('127.0.0.1', i))
            s.close()
            return i
        except socket.error:
            pass
    raise Exception('No local port available')


def _ssh_master_cmd(addr, user, command, local_key=None):
    """Exit or check ssh mux"""
    ssh_call = ['ssh', '-qNfL%d:127.0.0.1:12042' % find_port(addr, user),
                '-o', 'ControlPath=~/.ssh/unixpipe_%%r@%%h_%d' %
                       find_port(addr, user),
                '-O', command, '%s@%s' % (user, addr,)]

    if local_key:
        ssh_call.insert(1, local_key)
        ssh_call.insert(1, '-i')

    return subprocess.call(ssh_call)


def is_alive(addr, user):
    """Check wether a tunnel is alive"""
    return _ssh_master_cmd(addr, user, 'check') == 0


def setup(addr, user, remote_path, local_key=None):
    """Setup the tunnel"""
    port = find_port(addr, user)

    if not port or not is_alive(addr, user):
        port = new_port()

        scp(addr, user, __file__, '~/unixpipe', local_key)

        ssh_call = ['ssh', '-fL%d:127.0.0.1:12042' % port,
                    '-o', 'ExitOnForwardFailure=yes',
                    '-o', 'ControlPath=~/.ssh/unixpipe_%%r@%%h_%d' % port,
                    '-o', 'ControlMaster=auto',
                    '%s@%s' % (user, addr,), 'python', '~/unixpipe',
                    'server', remote_path]
        if local_key:
            ssh_call.insert(1, local_key)
            ssh_call.insert(1, '-i')

        subprocess.call(ssh_call)
        # XXX Sleep is a bad way to wait for the tunnel endpoint
        time.sleep(1)

    return port


if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == 'server':
        tcp4_to_unix(12042, sys.argv[2])
